from collections import OrderedDict
from typing import Type

from django.db import transaction
from django.db.models import Model
from django.utils.translation import gettext_lazy as _
from rest_framework import permissions
from rest_framework.exceptions import ValidationError
from rest_framework.generics import GenericAPIView
from rest_framework.response import Response
from rest_framework.viewsets import ModelViewSet

from ansible_base.lib.utils.views.django_app_api import AnsibleBaseDjangoAppApiView
from ansible_base.lib.utils.views.permissions import try_add_oauth2_scope_permission
from ansible_base.rbac.api.permissions import RoleDefinitionPermissions
from ansible_base.rbac.api.serializers import (
    RoleDefinitionDetailSerializer,
    RoleDefinitionSerializer,
    RoleMetadataSerializer,
    RoleTeamAssignmentSerializer,
    RoleUserAssignmentSerializer,
)
from ansible_base.rbac.evaluations import has_super_permission
from ansible_base.rbac.models import RoleDefinition
from ansible_base.rbac.permission_registry import permission_registry
from ansible_base.rbac.policies import check_can_remove_assignment
from ansible_base.rbac.validators import check_locally_managed, permissions_allowed_for_role, system_roles_enabled


def list_combine_values(data: dict[Type[Model], list[str]]) -> list[str]:
    "Utility method to merge everything in .values() into a single list"
    ret = []
    for this_list in data.values():
        ret += this_list
    return ret


class RoleMetadataView(AnsibleBaseDjangoAppApiView, GenericAPIView):
    """General data about models and permissions tracked by django-ansible-base RBAC

    Information from this endpoint should be static given a server version.
    This reflects model definitions, registrations with the permission registry,
    and enablement of RBAC features in settings.

    allowed_permissions: Valid permissions for a role of a given content_type
    """

    permission_classes = try_add_oauth2_scope_permission([permissions.IsAuthenticated])
    serializer_class = RoleMetadataSerializer

    def get(self, request, format=None):
        data = OrderedDict()
        allowed_permissions = OrderedDict()

        all_models = sorted(permission_registry.all_registered_models, key=lambda cls: cls._meta.model_name)

        role_model_types = list(all_models)
        if system_roles_enabled():
            role_model_types += [None]
        for cls in role_model_types:
            if cls is None:
                cls_repr = 'system'
            else:
                cls_repr = f"{permission_registry.get_resource_prefix(cls)}.{cls._meta.model_name}"
            allowed_permissions[cls_repr] = []
            for codename in list_combine_values(permissions_allowed_for_role(cls)):
                perm = permission_registry.permission_qs.get(codename=codename)
                ct = permission_registry.content_type_model.objects.get_for_id(perm.content_type_id)
                perm_repr = f"{permission_registry.get_resource_prefix(ct.model_class())}.{codename}"
                allowed_permissions[cls_repr].append(perm_repr)

        data['allowed_permissions'] = allowed_permissions

        serializer = self.get_serializer(data)

        return Response(serializer.data)


class RoleDefinitionViewSet(AnsibleBaseDjangoAppApiView, ModelViewSet):
    """
    Role Definitions (roles) contain a list of permissions and can be used to
    assign those permissions to a user or team through the respective
    assignment endpoints.

    Custom roles can be created, modified, and deleted through this endpoint.
    System-managed roles are shown here, which cannot be edited or deleted,
    but can be assigned to users.
    """

    queryset = RoleDefinition.objects.prefetch_related('created_by', 'modified_by', 'content_type', 'permissions')
    serializer_class = RoleDefinitionSerializer
    permission_classes = try_add_oauth2_scope_permission([RoleDefinitionPermissions])

    def get_serializer_class(self):
        if self.action == 'update':
            return RoleDefinitionDetailSerializer
        return super().get_serializer_class()

    def _error_if_managed(self, instance):
        if instance.managed is True:
            raise ValidationError(_('Role is managed by the system'))

    def perform_update(self, serializer):
        self._error_if_managed(serializer.instance)
        return super().perform_update(serializer)

    def perform_destroy(self, instance):
        self._error_if_managed(instance)
        return super().perform_destroy(instance)


assignment_prefetch_base = ('content_object', 'content_type', 'role_definition', 'created_by', 'object_role')


class BaseAssignmentViewSet(AnsibleBaseDjangoAppApiView, ModelViewSet):
    permission_classes = try_add_oauth2_scope_permission([permissions.IsAuthenticated])
    # PUT and PATCH are not allowed because these are immutable
    http_method_names = ['get', 'post', 'head', 'options', 'delete']
    prefetch_related = ()

    def get_queryset(self):
        model = self.serializer_class.Meta.model
        return model.objects.prefetch_related(*self.prefetch_related, *assignment_prefetch_base)

    def filter_queryset(self, qs):
        model = self.serializer_class.Meta.model
        if has_super_permission(self.request.user, 'view'):
            new_qs = qs
        else:
            new_qs = model.visible_items(self.request.user, qs)
        return super().filter_queryset(new_qs)

    def perform_create(self, serializer):
        return super().perform_create(serializer)

    def perform_destroy(self, instance):
        check_can_remove_assignment(self.request.user, instance)
        check_locally_managed(instance.role_definition)

        if instance.content_type_id:
            with transaction.atomic():
                instance.role_definition.remove_permission(instance.actor, instance.content_object)
        else:
            with transaction.atomic():
                instance.role_definition.remove_global_permission(instance.actor)


class RoleTeamAssignmentViewSet(BaseAssignmentViewSet):
    """
    Use this endpoint to give a team permission to a resource or an organization.
    The needed data is the team, the role definition, and the object id.
    The object must be of the type specified in the role definition.
    The type given in the role definition and the provided object_id are used
    to look up the resource.

    After creation, the assignment cannot be edited, but can be deleted to
    remove those permissions.
    """

    serializer_class = RoleTeamAssignmentSerializer
    prefetch_related = ('team',)


class RoleUserAssignmentViewSet(BaseAssignmentViewSet):
    """
    Use this endpoint to give a user permission to a resource or an organization.
    The needed data is the user, the role definition, and the object id.
    The object must be of the type specified in the role definition.
    The type given in the role definition and the provided object_id are used
    to look up the resource.

    After creation, the assignment cannot be edited, but can be deleted to
    remove those permissions.
    """

    serializer_class = RoleUserAssignmentSerializer
    prefetch_related = ('user',)



# Generated by Django 4.2.8 on 2024-04-10 19:57

from django.conf import settings
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('dab_authentication', '0005_alter_authenticator_created_by_and_more'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='authenticatoruser',
            unique_together={('provider', 'uid', 'user')},
        ),
    ]



import hashlib
from datetime import datetime, timezone

import pytest
from oauthlib.common import generate_token

from ansible_base.lib.testing.fixtures import copy_fixture
from ansible_base.lib.utils.hashing import hash_string
from ansible_base.lib.utils.response import get_relative_url
from ansible_base.oauth2_provider.models import OAuth2AccessToken, OAuth2Application

__all__ = [  # noqa: F822, the additional _pat_X are unknown so we need this here
    'oauth2_application',
    'oauth2_application_password',
    'oauth2_admin_access_token',
    'oauth2_user_pat',
    'oauth2_user_pat_1',
    'oauth2_user_pat_2',
    'oauth2_user_pat_3',
    'oauth2_user_application_token',
    'oauth2_user_application_token_1',
    'oauth2_user_application_token_2',
    'oauth2_user_application_token_3',
]


@pytest.fixture
def oauth2_application(randname):
    """
    Creates an OAuth2 application with a random name and returns
    both the application and its client secret.
    """
    app = OAuth2Application(
        name=randname("OAuth2 Application"),
        description="Test OAuth2 Application",
        redirect_uris="https://example.com/callback",
        authorization_grant_type="authorization-code",
        client_type="confidential",
    )
    # Store this before it gets hashed
    secret = app.client_secret
    app.save()
    return (app, secret)


@pytest.fixture
def oauth2_application_password(randname):
    """
    Creates an OAuth2 application with a random name and returns
    both the application and its client secret.
    """
    app = OAuth2Application(
        name=randname("OAuth2 Application"),
        description="Test OAuth2 Application",
        redirect_uris="https://example.com/callback",
        authorization_grant_type="password",
        client_type="confidential",
    )
    # Store this before it gets hashed
    secret = app.client_secret
    app.save()
    return (app, secret)


@pytest.fixture
def oauth2_admin_access_token(oauth2_application, admin_api_client, admin_user):
    """
    3-tuple with (token object with hashed token, plaintext token, plaintext_refresh_token)
    """
    url = get_relative_url('token-list')
    response = admin_api_client.post(url, {'application': oauth2_application[0].pk})
    assert response.status_code == 201

    plaintext_token = response.data['token']
    plaintext_refresh_token = response.data['refresh_token']
    hashed_token = hash_string(plaintext_token, hasher=hashlib.sha256, algo="sha256")
    token = OAuth2AccessToken.objects.get(token=hashed_token)
    return (token, plaintext_token, plaintext_refresh_token)


@copy_fixture(copies=3)
@pytest.fixture
def oauth2_user_pat(user, randname):
    return OAuth2AccessToken.objects.get_or_create(
        user=user,
        description=randname("Personal Access Token for 'user'"),
        # This has to be timezone aware
        expires=datetime(2088, 1, 1, tzinfo=timezone.utc),
        token=generate_token(),
    )[0]


@copy_fixture(copies=3)
@pytest.fixture
def oauth2_user_application_token(user, randname, oauth2_application):
    return OAuth2AccessToken.objects.get_or_create(
        user=user,
        description=randname("Application Access Token for 'user'"),
        # This has to be timezone aware
        expires=datetime(2088, 1, 1, tzinfo=timezone.utc),
        token=generate_token(),
        application=oauth2_application[0],
    )[0]



import logging
from typing import Optional, Tuple

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import AbstractUser
from django.db.models import Q
from django.utils.translation import gettext as _
from social_core.exceptions import AuthException
from social_core.pipeline.user import get_username

from ansible_base.authentication.authenticator_plugins.utils import get_authenticator_class
from ansible_base.authentication.models import Authenticator, AuthenticatorUser
from ansible_base.authentication.social_auth import AuthenticatorStorage, AuthenticatorStrategy

logger = logging.getLogger('ansible_base.authentication.utils.authentication')


class FakeBackend:
    def setting(self, *args, **kwargs):
        return ["username", "email"]


def migrate_from_existing_authenticator(
    uid: str, alt_uid: Optional[str], authenticator: Authenticator, preferred_username: Optional[str] = None
) -> Optional[str]:
    """
    uid: the users uid.
    alt_uid: an optional alternative uid to use for looking up other accounts.
    authenticator: the authenticator that the user is currently authenticating with.
    preferred_username: the username that the authenticator wants to use for the user.

    Returns the username of the django account to user for the authenticated user, or None if no match was found.
    """

    # SAML puts prepends all the UIDs with IdP. Adding this to the search criteria will
    # allow us to find SAML accounts that match the UID.
    uid_filter = [uid, "IdP:" + uid]
    if alt_uid:
        uid_filter.append(alt_uid)

    migrate_users = list(
        AuthenticatorUser.objects.filter(
            uid__in=uid_filter,
            provider__auto_migrate_users_to=authenticator.pk,
        ).order_by("provider__order")
    )

    if len(migrate_users) == 0:
        return

    try:
        main_user = AuthenticatorUser.objects.get(uid=uid, provider=authenticator).user
    except AuthenticatorUser.DoesNotExist:
        main_user = migrate_users[0].user

    for migrate_user in migrate_users:
        provider = migrate_user.provider
        from_authenticator = get_authenticator_class(provider.type)(database_instance=provider)
        old_user = from_authenticator.move_authenticator_user_to(main_user, migrate_user)
        if old_user and not old_user.authenticator_users.exists():
            old_user.delete()
        elif old_user:
            logger.warning(f"{old_user.username} is still managed by other authenticators and cannot be deleted.")

    # Now that we've potentially cleaned up any old user accounts, lets see if we can
    # give the user their preferred_username as their username

    if preferred_username:
        if main_user.username != preferred_username and not get_user_model().objects.filter(username=preferred_username).exists():
            main_user.username = preferred_username
            main_user.save()

    return main_user.username


def get_local_username(user_details: dict) -> str:
    """
    Converts the username provided by the backend to one that doesn't conflict with users
    from other auth backends.
    """
    username = get_username(strategy=AuthenticatorStrategy(AuthenticatorStorage()), details=user_details, backend=FakeBackend())

    return username['username']


def check_system_username(uid: str) -> None:
    """
    Determine if a username is identical with SYSTEM_USERNAME
    Raise AuthException if system user attempts to login via an external authentication source
    """
    if uid.casefold() == settings.SYSTEM_USERNAME.casefold():
        logger.warning(f'{settings.SYSTEM_USERNAME} cannot log in from an authenticator!')
        raise AuthException(_('System user is not allowed to log in from external authentication sources.'))


def determine_username_from_uid_social(**kwargs) -> dict:
    # If you are troubleshooting login issues and getting to here ....
    # Make sure that your backend properly implements get_user_details per the base backend
    # See https://github.com/python-social-auth/social-core/blob/master/social_core/backends/base.py#L173
    # The return should be in the format:
    #
    # {
    #     'username': <username if any>,
    #     'email': <user email if any>,
    #     'fullname': <user full name if any>,
    #     'first_name': <user first name if any>,
    #     'last_name': <user last name if any>
    # }
    # If this data structure does not return the username, this method will fail
    selected_username = kwargs.get('details', {}).get('username', None)
    if not selected_username:
        raise AuthException(
            _('Unable to get associated username from details, expected entry "username". Full user details: %(details)s')
            % {'details': kwargs.get("details", None)}
        )

    authenticator = kwargs.get('backend')
    if not authenticator:
        raise AuthException(_('Unable to get backend from kwargs'))

    alt_uid = authenticator.get_alternative_uid(**kwargs)

    if migrated_username := migrate_from_existing_authenticator(
        uid=kwargs.get("uid"), alt_uid=alt_uid, authenticator=authenticator.database_instance, preferred_username=selected_username
    ):
        username = migrated_username
    else:
        username = determine_username_from_uid(selected_username, authenticator.database_instance)

    return {"username": username}


def determine_username_from_uid(uid: str = None, authenticator: Authenticator = None, alt_uid: Optional[str] = None) -> str:
    """
    Determine what the username for the User object will be from the given uid and authenticator
    This will take uid like "bella" and search for an AuthenticatorUser and return:
        bella - if there is no bella user in AuthenticatorUser
        bella<hash> - if there is already a bella user in AuthenticatorUser but its not from the given authenticator
        <User.username> - If there is already a user associated with bella for this authenticator (could be bella or bella<hash> or even something else)

    NOTE: This should not be called directly. This will either be called from:
             1. The social auth pipeline
             2. The get_or_create_authenticator_user method below
          With one exception of the local authenticator. This is because we can't allow local authenticator to have maps between a uid of timmy and
          a username of timmy<hash>.  This literally does not make sense for the local authenticator because the DB is its own source of truth.
    """
    try:
        check_system_username(uid)
    except AuthException as e:
        logger.warning(f"AuthException: {e}")
        raise

    # If we have an AuthenticatorUser with the exact uid and provider than we have a match
    exact_match = AuthenticatorUser.objects.filter(uid=uid, provider=authenticator)
    if exact_match.count() == 1:
        new_username = exact_match[0].user.username
        logger.info(f"Authenticator {authenticator.name} already authenticated {uid} as {new_username}")
        return new_username

    # We didn't get an exact match. If any other provider is using this uid our id will be uid<hash>
    if AuthenticatorUser.objects.filter(Q(uid=uid) | Q(user__username=uid)).count() != 0:
        # Some other provider is providing this username so we need to create our own username
        new_username = get_local_username({'username': uid})
        logger.info(
            f'Authenticator {authenticator.name} wants to authenticate {uid} but that'
            f' username is already in use by another authenticator,'
            f' the user from this authenticator will be {new_username}'
        )
        return new_username

    # We didn't have an exact match but no other provider is servicing this uid so lets return that for usage
    logger.info(f"Authenticator {authenticator.name} is able to authenticate user {uid} as {uid}")
    return uid


def get_or_create_authenticator_user(
    uid: str, authenticator: Authenticator, user_details: dict = dict, extra_data: dict = dict
) -> Tuple[Optional[AbstractUser], Optional[AuthenticatorUser], Optional[bool]]:
    """
    Create the user object in the database along with it's associated AuthenticatorUser class.
    In some cases, the user may already be created in the database.
    This should be called any non-social auth plugins.

    Inputs
    uid: The unique id that identifies the user (this comes from the source authenticator, e.g. github or ldap).
    user_details: Any details about the user from the source (first name, last name, email, etc)
    authenticator: The authenticator authenticating the user
    extra_data: Any additional information about the user provided by the source.
                For example, LDAP might return sn, location, phone_number, etc
    """
    try:
        check_system_username(uid)
    except AuthException as e:
        logger.warning(f"AuthException: {e}")
        raise

    if migrated_username := migrate_from_existing_authenticator(uid=uid, alt_uid=None, authenticator=authenticator, preferred_username=uid):
        username = migrated_username
    else:
        username = determine_username_from_uid(uid, authenticator)

    created = None
    try:
        # First see if we have an auth user and if so update it
        auth_user = AuthenticatorUser.objects.get(uid=uid, provider=authenticator)
        auth_user.extra_data = extra_data
        auth_user.save()
        created = False
    except AuthenticatorUser.DoesNotExist:
        # Ensure that this username is not already tied to another authenticator
        auth_user = AuthenticatorUser.objects.filter(user__username=username).first()
        if auth_user is not None:
            logger.error(
                f'Authenticator {authenticator.name} attempted to create an AuthenticatorUser for {username}'
                f' but that id is already tied to authenticator {auth_user.provider.name}'
            )
            return None, None, None

    # ensure the authenticator isn't trying to pass along a cheeky is_superuser in user_details
    details = {k: user_details.get(k, "") for k in ["first_name", "last_name", "email"]}
    # Create or get our user object
    local_user, user_created = get_user_model().objects.get_or_create(username=username, defaults=details)
    if user_created:
        logger.info(f"Authenticator {authenticator.name} created User {username}")

    if created is None:
        # Create or get the user object. The get shouldn't happen but just incase something snuck in on us
        auth_user, created = AuthenticatorUser.objects.get_or_create(
            user=local_user,
            uid=uid,
            provider=authenticator,
            defaults={'extra_data': extra_data},
        )
        if created:
            extra = ''
            if not user_created:
                extra = ' attaching to existing user'
            logger.debug(f"Authenticator {authenticator.name} created AuthenticatorUser for {username}{extra}")

    return local_user, auth_user, created



from .authenticator import Authenticator
from .authenticator_map import AuthenticatorMap
from .authenticator_user import AuthenticatorUser

__all__ = (
    'Authenticator',
    'AuthenticatorMap',
    'AuthenticatorUser',
)



from django.db import migrations

from ansible_base.oauth2_provider.migrations._utils import hash_tokens


class Migration(migrations.Migration):
    dependencies = [
        ("dab_oauth2_provider", "0004_alter_oauth2accesstoken_scope"),
    ]

    operations = [
        migrations.RunPython(hash_tokens),
    ]



from .dynaconf_helpers import (
    export,
    factory,
    load_dab_settings,
    load_envvars,
    load_python_file_with_injected_context,
    load_standard_settings_files,
    toggle_feature_flags,
    validate,
)

__all__ = [
    "export",
    "factory",
    "load_dab_settings",
    "load_envvars",
    "load_python_file_with_injected_context",
    "load_standard_settings_files",
    "toggle_feature_flags",
    "validate",
]



from django.db import models


class ImmutableModel(models.Model):
    """
    A save-once (immutable) base model. Simply blocks any save attempts after the first.
    """

    class Meta:
        abstract = True

    def save(self, *args, **kwargs):
        if self.pk:
            raise ValueError(f"{self.__class__.__name__} is immutable and cannot be modified.")

        return super().save(*args, **kwargs)



from typing import Any, Type, Union
from uuid import UUID

from django.apps import apps as django_apps
from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.exceptions import ImproperlyConfigured
from django.db.models import CharField, Model, OuterRef
from django.db.models.functions import Cast
from django.db.models.query import QuerySet

from ansible_base.lib.abstract_models.organization import AbstractOrganization
from ansible_base.lib.abstract_models.team import AbstractTeam


def get_model_from_settings(setting_name: str) -> Any:
    """
    Return the User model that is active in this project.
    """
    try:
        setting = getattr(settings, setting_name)
    except AttributeError:
        raise ImproperlyConfigured(f"{setting_name} is not defined in settings.")
    try:
        return django_apps.get_model(setting, require_ready=False)
    except ValueError:
        raise ImproperlyConfigured(f"{setting_name} must be of the form 'app_label.model_name'")
    except LookupError:
        raise ImproperlyConfigured(f"{setting_name} refers to model '{setting}' that has not been installed")


def get_team_model() -> Type[AbstractTeam]:
    return get_model_from_settings('ANSIBLE_BASE_TEAM_MODEL')


def get_organization_model() -> Type[AbstractOrganization]:
    return get_model_from_settings('ANSIBLE_BASE_ORGANIZATION_MODEL')


def get_object_by_ansible_id(qs: QuerySet, ansible_id: Union[str, UUID], annotate_as: str = 'ansible_id_for_filter') -> Model:
    resource_cls = django_apps.get_model('dab_resource_registry', 'Resource')
    content_type_cls = django_apps.get_model('contenttypes', 'ContentType')
    cls = qs.model
    ct = content_type_cls.objects.get_for_model(cls)
    pk_field_name = cls._meta.pk.name
    pk_reference = Cast(OuterRef(pk_field_name), output_field=CharField())
    resource_qs = resource_cls.objects.filter(object_id=pk_reference, content_type=ct).values('ansible_id')
    return qs.annotate(**{annotate_as: resource_qs}).get(**{annotate_as: ansible_id})


def get_user_by_ansible_id(ansible_id: Union[str, UUID], annotate_as: str = 'ansible_id_for_filter') -> Model:
    return get_object_by_ansible_id(get_user_model().objects.all(), ansible_id, annotate_as=annotate_as)



import logging
from collections import OrderedDict

from django.conf import settings
from django.contrib.auth.models import AbstractUser
from django.db import models
from django.urls.exceptions import NoReverseMatch
from django.utils.translation import gettext_lazy as _
from inflection import underscore

from ansible_base.lib.abstract_models.immutable import ImmutableModel
from ansible_base.lib.utils.encryption import ansible_encryption
from ansible_base.lib.utils.models import current_user_or_system_user, is_system_user
from ansible_base.lib.utils.response import get_relative_url

logger = logging.getLogger('ansible_base.lib.abstract_models.common')


def get_cls_view_basename(cls):
    # This gives the expected base Django view name of cls
    if hasattr(cls, 'router_basename'):
        return cls.router_basename
    return underscore(cls.__name__)


def get_url_for_object(obj, request=None, pk=None):
    # get_absolute_url mainly exists to support AWX
    if hasattr(obj, 'get_absolute_url'):
        return obj.get_absolute_url()

    basename = get_cls_view_basename(obj.__class__)

    try:
        return get_relative_url(f'{basename}-detail', kwargs={'pk': pk or obj.pk})
    except NoReverseMatch:
        logger.debug(f"Tried to reverse {basename}-detail for model {obj.__class__.__name__} but said view is not defined")
        return ''


class ModifiableModel(models.Model):
    class Meta:
        abstract = True

    modified = models.DateTimeField(
        editable=False,
        help_text=_("The date/time this resource was created."),
        auto_now=True,
    )
    modified_by = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        related_name='%(app_label)s_%(class)s_modified+',
        default=None,
        null=True,
        editable=False,
        on_delete=models.SET_NULL,
        help_text=_("The user who last modified this resource."),
    )

    def save(self, *args, **kwargs):
        '''
        This save function will provide the following features automatically.
          * It will automatically add the modified fields for changing items

        There are several edge cases to be aware of:
          * If the user is logging in and their "last_login" field is getting
            changed (in their user model instance), then we do not update
            "modified_by". (This is so they can see if someone changed their
            user object while they were away).
          * If no fields in update_fields are "editable", then we don't update
            "modified_by".
          * "modified" is updated in every case, since it's an auto_now
            timestamp.
        '''
        has_update_fields = kwargs.get('update_fields') is not None
        update_fields = kwargs.get('update_fields', [])
        if update_fields is None:
            update_fields = []
        else:
            update_fields = list(update_fields)
        is_user_logging_in = isinstance(self, AbstractUser) and update_fields == ['last_login']
        has_editable_field = any(self._meta.get_field(field).editable for field in update_fields)

        if 'modified_by' in update_fields:
            # Explicit update of modified_by, don't mess with it
            pass
        elif is_user_logging_in:
            # User is logging in, only last_login is changing, don't update modified_by
            pass
        elif has_update_fields and not has_editable_field:
            # No editable fields are changing, don't update modified_by
            pass
        elif not self.pk and is_system_user(self):
            # If we are building the system user we can't set the modified_by or it will error (because it would be system which isn't built yet)
            pass
        else:
            self.modified_by = current_user_or_system_user()
            update_fields.append('modified_by')
            if has_update_fields:
                kwargs['update_fields'] = update_fields

        return super().save(*args, **kwargs)


class CreatableModel(models.Model):
    class Meta:
        abstract = True

    created = models.DateTimeField(
        editable=False,
        help_text=_("The date/time this resource was created."),
        auto_now_add=True,
    )
    created_by = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        related_name='%(app_label)s_%(class)s_created+',
        default=None,
        null=True,
        editable=False,
        on_delete=models.SET_NULL,
        help_text=_("The user who created this resource."),
    )

    def save(self, *args, **kwargs):
        '''
        This save function will provide the following features automatically.
          * It will automatically add a created_by fields for new items
        '''

        if not self.pk and not is_system_user(self):
            if self.created_by is None:
                self.created_by = current_user_or_system_user()

        return super().save(*args, **kwargs)


class AbstractCommonModel(models.Model):
    # Any field marked as encrypted will automatically be stored in an encrypted fashion
    encrypted_fields = []
    # Any field set in here will not be used in the views
    ignore_relations = []

    class Meta:
        abstract = True

    def save(self, *args, **kwargs):
        '''
        This save function will provide the following features automatically.
          * It will automatically encrypt any fields in the classes `encrypt_fields` property
        '''
        # Encrypt any fields
        for field in self.encrypted_fields:
            field_value = getattr(self, field, None)
            setattr(self, field, ansible_encryption.encrypt_string(field_value))

        return super().save(*args, **kwargs)

    @classmethod
    def from_db(self, db, field_names, values):
        instance = super().from_db(db, field_names, values)

        for field in self.encrypted_fields:
            field_value = getattr(instance, field, None)
            setattr(instance, field, ansible_encryption.decrypt_string(field_value))

        return instance

    def get_summary_fields(self):
        response = {}
        for field in self._meta.fields:
            if isinstance(field, models.ForeignObject) and getattr(self, field.name):
                # ignore relations on inherited django models
                if field.name.endswith("_ptr") or (field.name in self.ignore_relations):
                    continue
                if hasattr(getattr(self, field.name), 'summary_fields'):
                    response[field.name] = getattr(self, field.name).summary_fields()
        return response

    def related_fields(self, request):
        response = {}
        # See docs/lib/default_models.md
        # Automatically add all of the ForeignKeys for the model as related fields
        for field in self._meta.concrete_fields:
            if field.name in self.ignore_relations:
                continue
            # ignore relations on inherited django models
            if not isinstance(field, models.ForeignKey) or field.name.endswith("_ptr"):
                continue

            if obj := getattr(self, field.name):
                if related_url := get_url_for_object(obj):
                    response[field.name] = related_url

        basename = get_cls_view_basename(self.__class__)

        # Add any reverse relations required
        missing_relations = []
        for relation in self._meta.related_objects + self._meta.many_to_many:
            field_name = relation.name
            # obey the model ignore list
            if field_name in self.ignore_relations:
                continue
            reverse_view = f"{basename}-{field_name}-list"
            try:
                response[field_name] = get_relative_url(reverse_view, kwargs={'pk': self.pk})
            except NoReverseMatch:
                missing_relations.append(reverse_view)

        if missing_relations and settings.DEBUG:
            logger.error(f"Wanted to add {', '.join(missing_relations)} for {self.__class__} but view was missing")

        if hasattr(self, 'extra_related_fields'):
            response.update(self.extra_related_fields(request))

        sorted_response = OrderedDict()
        sorted_keys = list(response.keys())
        sorted_keys.sort()
        for key in sorted_keys:
            sorted_response[key] = response[key]
        return sorted_response

    def summary_fields(self):
        response = {}
        response['id'] = self.id
        return response


class CommonModel(AbstractCommonModel, CreatableModel, ModifiableModel):
    """
    A base model for most django-ansible-base apps to extend from.
    Includes fields for tracking creation and last-modification metadata.
    """

    class Meta:
        abstract = True


class NamedCommonModel(CommonModel):
    class Meta:
        abstract = True

    name = models.CharField(
        max_length=512,
        help_text=_("The name of this resource."),
    )

    def summary_fields(self):
        res = super().summary_fields()
        res['name'] = self.name
        return res

    def __str__(self):
        return self.name


class UniqueNamedCommonModel(CommonModel):
    class Meta:
        abstract = True

    name = models.CharField(
        max_length=512,
        unique=True,
        help_text=_("The name of this resource."),
    )

    def summary_fields(self):
        res = super().summary_fields()
        res['name'] = self.name
        return res

    def __str__(self):
        return self.name


class ImmutableCommonModel(ImmutableModel, AbstractCommonModel, CreatableModel):
    """
    A save-once (immutable) base model.
    Functionally similar to CommonModel, but does not allow modification of the object after creation
    and does not include the modified/modifed_by fields.
    """

    class Meta:
        abstract = True



import logging

logger = logging.getLogger('ansible_base.resource_registry.utils.sso_provider')

try:
    from social_django.utils import load_strategy
except ImportError:
    load_strategy = None


def get_sso_provider_server(backend_name: str, uid: str):
    """
    Returns the hostname for the SSO server that authenticated this user.
    """

    # If this function fails for any reason it will cause the resources API
    # to break. There are too many potential permutations of SSO configurations
    # to handle all of failure modes here. This function will try its best to get
    # the server url, but if anything fails, we should just return None rather than
    # break. There are a few backends that we do care about, and these will be
    # tested to make sure that they work, but it doesn't matter if some just
    # return None.
    try:
        if load_strategy is None:
            logger.debug("Cannot get SSO backend. 'social_core' is not installed.")
            return (None, uid)

        social_strat = load_strategy()

        try:
            backend = social_strat.get_backend(backend_name)
        except TypeError:
            return (None, uid)

        if hasattr(backend, "get_idp"):
            idp, real_uid = uid.split(":", maxsplit=1)
            return (backend.get_idp(idp).entity_id, real_uid)

        elif hasattr(backend, "authorization_url"):
            return (backend.authorization_url(), uid)
        else:
            return (None, uid)
    except Exception as e:  # noqa E722
        logger.warning(f"Failed to parse server url from {backend_name}, {e}")
        return (None, uid)



from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import FieldError
from django.db.models import Q
from rest_framework.exceptions import ParseError
from rest_framework.filters import BaseFilterBackend

from ansible_base.lib.utils.models import get_all_field_names, get_type_for_model


class TypeFilterBackend(BaseFilterBackend):
    """
    Filter on type field now returned with all objects.
    """

    def filter_queryset(self, request, queryset, view):
        try:
            types = None
            for key, value in request.query_params.items():
                if key == 'type':
                    if ',' in value:
                        types = value.split(',')
                    else:
                        types = (value,)
            if types:
                types_map = {}
                # TODO: We need to revisit this out since we don't have a main app
                for ct in ContentType.objects.filter(Q(app_label='main') | Q(app_label='auth', model='user')):
                    ct_model = ct.model_class()
                    if not ct_model:
                        continue
                    ct_type = get_type_for_model(ct_model)
                    types_map[ct_type] = ct.pk
                model = queryset.model
                model_type = get_type_for_model(model)
                if 'polymorphic_ctype' in get_all_field_names(model):
                    types_pks = set([v for k, v in types_map.items() if k in types])
                    queryset = queryset.filter(polymorphic_ctype_id__in=types_pks)
                elif model_type not in types:
                    queryset = queryset.none()
            return queryset
        except FieldError as e:
            # Return a 400 for invalid field names.
            raise ParseError(*e.args)



# Generated by Django 4.2.6 on 2024-03-01 18:41

from django.db import migrations, models

import ansible_base.resource_registry.models.resource


def reset_service_id(apps, schema_editor):
    ServiceID = apps.get_model("dab_resource_registry", "ServiceID")
    Resource = apps.get_model("dab_resource_registry", "Resource")
    Resource.objects.all().update(service_id=ServiceID.objects.first().id)


def revert_post_migrate(apps, schema_editor):
    "Resources are created post-migrate, and prior model is incompatible anyway, so delete them"
    Resource = apps.get_model("dab_resource_registry", "Resource")
    Resource.objects.all().delete()


class Migration(migrations.Migration):

    dependencies = [
        ('dab_resource_registry', '0001_initial'),
    ]

    operations = [
        migrations.RenameField(
            model_name='resource',
            old_name='resource_id',
            new_name='ansible_id',
        ),
        migrations.AlterField(
            model_name='resource',
            name='service_id',
            field=models.CharField(max_length=38),
        ),
        migrations.RunPython(
            code=reset_service_id,
            reverse_code=revert_post_migrate
        ),
        migrations.AlterField(
            model_name='resource',
            name='service_id',
            field=models.UUIDField(default=ansible_base.resource_registry.models.resource.service_id),
        ),
    ]



from django.apps import AppConfig

from ansible_base.api_documentation.customizations import apply_authentication_customizations


class ApiDocumentationConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'ansible_base.api_documentation'
    label = 'dab_api_documentation'

    def ready(self):
        from django.conf import settings

        if 'ansible_base.authentication' in settings.INSTALLED_APPS:
            apply_authentication_customizations()



import hashlib
import json
from typing import Callable, Optional, Type

from django.db.models import Model
from rest_framework.serializers import Serializer


def hash_serializer_data(instance: Model, serializer: Type[Serializer], field: Optional[str] = None, hasher: Callable = hashlib.sha256):
    """
    Takes an instance, serialize it and take the .data or the specified field
    as input for the hasher function.
    """
    serialized_data = serializer(instance).data
    if field:
        serialized_data = serialized_data[field]
    metadata_json = json.dumps(serialized_data, sort_keys=True).encode("utf-8")
    return hasher(metadata_json).hexdigest()


def hash_string(inp: str, hasher: Callable = hashlib.sha256, algo=""):
    """
    Takes a string and hashes it with the given hasher function.
    If algo is given, it is prepended to the hash between dollar signs ($)
    before the hash is returned.

    NOTE: There is no salt or pepper here, so this is not secure for passwords.
    It is, however, useful for *random* strings like tokens, that need to be secured.
    """
    hash = hasher(inp.encode("utf-8")).hexdigest()
    if algo:
        return f"${algo}${hash}"
    return hash



from django import template

from ansible_base.lib.utils import requests as dab_requests

register = template.Library()


@register.simple_tag
def is_proxied_request():
    return dab_requests.is_proxied_request()



import functools
import time
from collections.abc import Callable
from logging import Logger, getLogger

__all__ = ['log_excess_runtime']


# This allows for configuring all logs from this to go to a particular destination
# this is not a child of ansible_base so it can be configured separately, and default to not configured
dab_logger = getLogger('dab_runtime_logger')


DEFAULT_MSG = 'Running {name} took {delta:.2f}s'


def log_excess_runtime(ext_logger: Logger, cutoff: float = 5.0, debug_cutoff: float = 2.0, msg: str = DEFAULT_MSG, add_log_data: bool = False):
    """Utility to write logs to the passed-in logger if the function it decorates takes too long.

    Call this with the configuration options to return a decorator."""

    def log_excess_runtime_decorator(func: Callable):
        """Instantiated decorator for the DAB excess runtime logger utility"""

        @functools.wraps(func)
        def _new_func(*args, **kwargs):
            log_data = {'name': repr(func.__name__)}
            start_time = time.time()

            if add_log_data:
                return_value = func(*args, log_data=log_data, **kwargs)
            else:
                return_value = func(*args, **kwargs)

            log_data['delta'] = time.time() - start_time

            formatted_msg = msg.format(**log_data)

            for logger in (dab_logger, ext_logger):
                if log_data['delta'] > cutoff:
                    logger.info(formatted_msg)
                elif log_data['delta'] > debug_cutoff:
                    logger.debug(formatted_msg)

            return return_value

        return _new_func

    return log_excess_runtime_decorator



from typing import Iterable, Optional, Type, Union

from django.apps import AppConfig
from django.conf import settings
from django.core.checks import CheckMessage, Debug, Error, Warning
from rest_framework.permissions import AllowAny, OperandHolder, OperationHolderMixin, SingleOperandHolder
from rest_framework.views import APIView

from ansible_base.lib.utils.views.urls import get_api_view_functions
from ansible_base.oauth2_provider.permissions import OAuth2ScopePermission


class OAuth2ScopePermissionCheck:
    """
    Class containing logic for checking view classes for the
    OAuth2ScopePermission permission_class, and aggregating
    CheckMessage's for django system checks.

    :param ignore_list: List of python import path strings for view classes exempt from the check logic.
    :type ignore_list: list
    """

    def __init__(self, ignore_list: Iterable[str], generate_check_messages=True):
        self.messages: list[CheckMessage] = []
        self.current_view: Optional[Type[APIView]] = None
        self.ignore_list = ignore_list
        self.generate_check_messages = generate_check_messages

    def check_message(self, message: CheckMessage):
        if self.generate_check_messages:
            self.messages.append(message)

    # These are all warning or error conditions, this function is mostly saying to not invert OAuth2ScopePermissions
    # Returns False.
    def process_single_operand_holder(self, operand_holder: SingleOperandHolder) -> bool:
        # The only unary operand for permission classes provided by rest_framework is ~ (not)
        if self.parse_permission_class(operand_holder.op1_class):
            self.check_message(
                Warning(
                    "~ (not) operand used on OAuth2ScopePermission, probably a bad idea.",
                    id="ansible_base.oauth2_provider.W001",
                    obj=self.current_view,
                )
            )

        return False

    def process_operand_holder(self, operand_holder: OperandHolder) -> bool:
        return self.parse_permission_class(operand_holder.op1_class) or self.parse_permission_class(operand_holder.op2_class)

    # Check if permission class is present in nested operands
    # Sort of recursive? Reasonably this should not be an issue, so long as we don't recurse on an unknown OperationHolderMixin type
    def parse_permission_class(self, cls: Union[Type[OperationHolderMixin], OperationHolderMixin]) -> bool:
        # First, most likely case, we're dealing with a BasePermission subclass.
        if cls is OAuth2ScopePermission:
            return True
        elif isinstance(cls, SingleOperandHolder):
            # Warning or Error case: Will not accept OAuth2 permission nested in NOT
            return self.process_single_operand_holder(cls)
        elif isinstance(cls, OperandHolder):
            return self.process_operand_holder(cls)
        return False

    def check_view(self, view_class: Type[APIView]) -> bool:
        """
        Primary function of the OAuth2ScopePermissionCheck.

        Checks if OAuth2ScopePermission is present on the supplied view's
        permission_classes; ignores classes that are not APIViews, or that are
        in the ignore_list.

        Appends CheckMessages to self.messages as a side effect.

        :param view_class: django View class or rest_framework ApiView class

        :return: True if view_class uses the OAuth2ScopePermission permission
            class, or has some mitigating circumstance that prohibits it, such as
            view_class not using permission classes, or its import path being in
            self.ignore_list; returns False otherwise.
        """
        if f"{view_class.__module__}.{view_class.__name__}" in self.ignore_list:
            self.check_message(
                Debug(
                    "View class in the ignore list. Ignoring.",
                    obj=view_class,
                    id="ansible_base.oauth2_provider.D03",
                )
            )
            return True

        self.current_view = view_class

        for permission_class in getattr(self.current_view, "permission_classes", []):
            if self.parse_permission_class(permission_class):
                self.check_message(
                    Debug(
                        "Found OAuth2ScopePermission permission_class",
                        obj=self.current_view,
                        id="ansible_base.oauth2_provider.D02",
                    )
                )
                return True

        if not self.current_view.permission_classes or AllowAny in self.current_view.permission_classes:
            self.check_message(
                Debug(
                    "View object is fully permissive, OAuth2ScopePermission is not required",
                    obj=self.current_view,
                    id="ansible_base.oauth2_provider.D04",
                )
            )
            return True

        # if we went though the whole loop without finding a valid permission_class, raise an error
        self.check_message(
            Error(
                "View class has no valid usage of OAuth2ScopePermission",
                obj=self.current_view,
                id="ansible_base.oauth2_provider.E002",
            )
        )
        return False


def view_in_app_configs(view_class: type, app_configs: Optional[list[AppConfig]]) -> bool:
    if app_configs:
        for app_config in app_configs:
            if view_class.__module__.startswith(app_config.name):
                return True
        return False
    return True


def oauth2_permission_scope_check(app_configs: Optional[list[AppConfig]], **kwargs) -> list[CheckMessage]:
    """
    Check for OAuth2ScopePermission permission class on all enabled views.

    Ignore views in the ANSIBLE_BASE_OAUTH2_PROVIDER_PERMISSIONS_CHECK_IGNORED_VIEWS setting
    """
    ignore_list = set(
        getattr(settings, "ANSIBLE_BASE_OAUTH2_PROVIDER_PERMISSIONS_CHECK_DEFAULT_IGNORED_VIEWS", [])
        + getattr(settings, "ANSIBLE_BASE_OAUTH2_PROVIDER_PERMISSIONS_CHECK_IGNORED_VIEWS", [])
    )

    check = OAuth2ScopePermissionCheck(ignore_list)

    view_functions = get_api_view_functions()
    for view in view_functions:
        # Only run checks on included apps (or all if app_configs is None)
        if view_in_app_configs(view, app_configs):
            check.check_view(view)

    return check.messages



import logging

from rest_framework import status
from rest_framework.response import Response
from rest_framework.viewsets import ModelViewSet

from ansible_base.authentication.models import Authenticator, AuthenticatorUser
from ansible_base.authentication.serializers import AuthenticatorSerializer
from ansible_base.lib.utils.views.django_app_api import AnsibleBaseDjangoAppApiView

logger = logging.getLogger('ansible_base.authentication.views.authenticator')


class AuthenticatorViewSet(AnsibleBaseDjangoAppApiView, ModelViewSet):
    """
    API endpoint that allows authenticators to be viewed or edited.
    """

    queryset = Authenticator.objects.all()
    serializer_class = AuthenticatorSerializer

    def get_serializer(self, *args, **kwargs):
        # Return an instanced serializer if one exists for OPTIONS requests.
        # This is because the AuthenticatorSerializer uses our ImmutableFieldsMixin,
        # which dynamically changes serializer fields.
        if self.action == "metadata":
            instance = kwargs.get("instance", None)
            if not instance and len(args) == 0:
                lookup_url_kwarg = self.lookup_url_kwarg or self.lookup_field
                if lookup_url_kwarg in self.kwargs:
                    # set instance to any truthy value, the serializer should not need more details to initialize its field information.
                    kwargs["instance"] = "object"

        return super().get_serializer(*args, **kwargs)

    def destroy(self, request, *args, **kwargs):
        instance = self.get_object()

        # check if there is at least one other enabled authenticator...
        if instance.enabled and instance.is_last_enabled:
            logger.warning("Preventing deletion of last enabled authenticator")
            return Response(status=status.HTTP_400_BAD_REQUEST, data={"details": "Authenticator cannot be deleted, as no authenticators would be enabled"})

        users_exist = AuthenticatorUser.objects.filter(provider_id=instance.slug).exists()
        if users_exist:
            logger.info("Found existing users from the authenticator")
            return Response(
                status=status.HTTP_409_CONFLICT, data={"details": "Authenticator cannot be deleted, as users from the authenticator exist in the system"}
            )
        else:
            logger.info(f"Deleting authenticator with ID={instance.id}")
            return super().destroy(request, *args, **kwargs)



import copy
import logging
import os
from typing import Any, Awaitable, Dict, Optional, Union
from urllib.parse import parse_qs, urlparse

from django.core.exceptions import ImproperlyConfigured
from django.utils.translation import gettext as _
from django_redis.client import DefaultClient
from redis import Redis
from redis.cluster import ClusterNode, RedisCluster
from redis.exceptions import NoPermissionError, RedisClusterException

from ansible_base.lib.constants import STATUS_DEGRADED, STATUS_FAILED, STATUS_GOOD

logger = logging.getLogger('ansible_base.lib.redis.client')

_DEFAULT_STATUS_TIMEOUT_SEC = 4
_REDIS_CLUSTER_OK_STATUS = 'ok'

_INVALID_STANDALONE_OPTIONS = ['cluster_error_retry_attempts']

# This will eventually be in the Redis package under redis.typing later
ResponseT = Union[Awaitable, Any]


class DABRedisRespectACLFlushMixin:
    """
    FLUSHDB is called when we call django.cache.clear()

    FLUSHDB does NOT respect Redis ACL on keys. It flushes all keys in a database.
    You think your keys are protected because the user you are running FLUSHDB as
    has write-only access to keys with the pattern gateway* or even read-only access
    ... WRONG, you aren't safe!

    And if you are thinking of using Redis' command re-write feature to sweep this
    problem under the rug then god help the developer or support person that comes after you.

    At some point in the future we may also want to override flushall as well
    """

    def flushdb(self, asynchronous: Optional[bool] = None, **kwargs) -> ResponseT:
        if asynchronous is not None:
            logger.warning("DABRedis clients implement an ACL friendly FLUSHDB which can not be async")

        if self.__class__ == DABRedisCluster:
            all_keys = self.keys("*", target_nodes=self.ALL_NODES)
        elif self.__class__ == DABRedis:
            all_keys = self.keys("*")
        else:
            raise ImproperlyConfigured(f"Got an inappropriate type of client {self.__class__}")

        if all_keys:
            # Only attempt to delete keys if we got some otherwise we will get an exception from the delete function
            try:
                self.delete(*all_keys)
            except NoPermissionError:
                pass


# We are going to build our own cluster class to override the mget function
# In a redis cluster, keys might not be in the same slot and this will throw off mget.
# Instead, we are going to try and use mget and then, if we get the slot error, we will try the mget_nonatomic to make it work
class DABRedisCluster(DABRedisRespectACLFlushMixin, RedisCluster):
    mode = 'cluster'

    def mget(self, *args, **kwargs):
        try:
            return super().mget(*args, **kwargs)
        except RedisClusterException as e:
            if 'MGET - all keys must map to the same key slot' in str(e):
                return super().mget_nonatomic(*args, **kwargs)
            raise


class DABRedis(DABRedisRespectACLFlushMixin, Redis):
    mode = 'standalone'


class RedisClient(DefaultClient):
    """
    Get a redis_client for the django cache
    """

    def _get_client_args(self):
        return self._params.get('OPTIONS', {}).get('CLIENT_CLASS_KWARGS', {})

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        connection_kwargs = self._get_client_args()
        self.mode = connection_kwargs.get('mode', 'standalone')
        self.redis_hosts = connection_kwargs.get('redis_hosts', '')

    def connect(self, index: int = 0) -> Union[DABRedis, DABRedisCluster]:
        """
        Given a connection index, returns a new raw redis client/connection
        instance. Index is used for replication setups and indicates that
        connection string should be used. In normal setups, index is 0.
        """

        # Make a deep copy of the CLIENT_CLASS_KWARGS so we don't accidentally modify the actual settings
        connection_args = copy.deepcopy(self._get_client_args())

        return get_redis_client(url=self._server[index], **connection_args)


class RedisClientGetter:
    def _redis_parse_url(self) -> None:
        if self.url == '':
            # If there is no URL we have nothing to do
            return

        # If we can't parse this just let it raise because other things will fail anyway
        parsed_url = urlparse(self.url)
        if parsed_url.scheme in ['file', 'unix']:
            # Attempt to attach to a socket if its a file or unix scheme
            self.connection_settings['unix_socket_path'] = parsed_url.path
        elif parsed_url.scheme in ['redis', 'rediss']:
            # Extract information from a rediss url
            for arg_name, parse_name in [('host', 'hostname'), ('port', 'port'), ('username', 'username'), ('password', 'password')]:
                attribute = getattr(parsed_url, parse_name, None)
                if attribute:
                    self.connection_settings[arg_name] = attribute

            if parsed_url.scheme == 'rediss' and self.connection_settings.get('ssl', None) is None:
                logger.info('Implicitly setting ssl in kwargs because URL was rediss')
                self.connection_settings['ssl'] = True

            if self.connection_settings.get('ssl') and parsed_url.scheme != 'rediss':
                self.url = self.url.replace('redis://', 'rediss://')
                logger.info("Altering redis URL to be rediss because tls was specified in kwargs")
        else:
            raise ImproperlyConfigured(_('Invalid redis URL, can only have the scheme file, unix, redis or rediss URLs, got {}').format(self.url))

        # Add the DB from the URL (if passed)
        try:
            self.connection_settings['db'] = int(parsed_url.path.split('/')[1])
        except (IndexError, ValueError):
            pass

        # Add any additional query params from the URL as kwargs
        for key, value in parse_qs(parsed_url.query).items():
            self.connection_settings[key] = value[-1]

    def _get_hosts(self) -> None:
        if not self.redis_hosts or self.redis_hosts == '':
            return

        self.connection_settings.pop('host', None)
        self.connection_settings.pop('port', None)
        self.connection_settings['startup_nodes'] = []

        # Make sure we have a string for redis_hosts
        try:
            if not isinstance(self.redis_hosts, str):
                logger.error(f"Specified redis_hosts is not a string, got: {self.redis_hosts}")
                # Since we didn't get a string we can't test any further, raise an exception here
                raise ValueError()

            had_host_errors = False
            host_ports = self.redis_hosts.split(',')
            for host_port in host_ports:
                try:
                    node, port_string = host_port.split(':')
                except ValueError:
                    logger.error(f"Specified cluster_host {host_port} is not valid; it needs to be in the format <host>:<port>")
                    had_host_errors = True
                    continue

                # Make sure we have an int for the port
                try:
                    port = int(port_string)
                except ValueError:
                    logger.error(f'Specified port on {host_port} is not an int')
                    had_host_errors = True
                    continue

                self.connection_settings['startup_nodes'].append(ClusterNode(node, port))

            if had_host_errors:
                raise ValueError()

        except ValueError:
            raise ImproperlyConfigured(_('Unable to parse redis_hosts, see logs for more details'))

    def __init__(self, *args, **kwargs):
        self.url = ''

    def get_client(self, url: str = '', **kwargs) -> Union[DABRedis, DABRedisCluster]:
        # remove our settings which are invalid to the parent classes
        self.mode = kwargs.pop('mode', 'standalone')
        self.redis_hosts = kwargs.pop('redis_hosts', None)

        self.connection_settings = kwargs
        self.url = url
        self._redis_parse_url()

        for file_setting in ['ssl_certfile', 'ssl_keyfile', 'ssl_ca_certs']:
            file = self.connection_settings.get(file_setting, None)
            if file == '':
                # Underlying libraries inspect these settings like `if ssl_ca_certs is not None`
                # However, we allow people to unset these by setting env vars as ""
                # So if we get a '' or None we are going to just remove the setting so that the underlying library does not try to validate a file named ''
                self.connection_settings.pop(file_setting)
            elif file is not None and self.connection_settings.get('ssl', None) and not os.access(file, os.R_OK):
                raise ImproperlyConfigured(_('Unable to read file {} from setting {}').format(file, file_setting))

        # Connect to either a cluster or a standalone redis
        if self.mode == 'cluster':
            logger.debug("Connecting to Redis clustered")
            self._get_hosts()
            return DABRedisCluster(**self.connection_settings)
        elif self.mode == 'standalone':
            for setting in _INVALID_STANDALONE_OPTIONS:
                if setting in self.connection_settings:
                    logger.info(f"Removing setting {setting} from connection settings because its invalid for standalone mode")
                    self.connection_settings.pop(setting)
            logger.debug("Connecting to Redis standalone")
            return DABRedis(**self.connection_settings)
        else:
            raise ImproperlyConfigured(_("mode must be either one of ['cluster', 'standalone'] got {}").format(self.mode))


def get_redis_client(url: str = '', **kwargs) -> Union[DABRedis, DABRedisCluster]:
    """
    Get a raw redis client based on a combination of url and kwargs
    The URL can contain things like the db and other params which will be converted into kwargs for the underlying redis client
    Or parameters for the underlying redis client can be specified directly in kwargs
    This will return a DABRedisCluster based on the kwargs "clustered" otherwise it will return a regular DABRedis client
    If clustered is specified this function also expects the setting "redis_hosts" as a string of host:port,host:port....
    """
    client_getter = RedisClientGetter()
    return client_getter.get_client(url, **kwargs)


def get_redis_status(url: str = '', timeout: int = _DEFAULT_STATUS_TIMEOUT_SEC, **kwargs) -> Dict:
    for setting in ['socket_timeout', 'socket_connect_timeout']:
        if setting not in kwargs:
            kwargs[setting] = timeout
    if 'socket_keepalive' not in kwargs:
        kwargs['socket_keepalive'] = True

    kwargs.pop('retry', None)

    response = {
        'mode': 'Unknown',
        'status': 'Unknown',
    }

    try:
        redis_client = get_redis_client(url, **kwargs)
        response['mode'] = redis_client.mode
        response['status'] = STATUS_GOOD
        if redis_client.mode == 'standalone':
            response['ping'] = redis_client.execute_command('PING')
        elif redis_client.mode == 'cluster':
            response['cluster_info'] = redis_client.cluster_info()
            response['cluster_nodes'] = redis_client.cluster_nodes()
            response['status'] = determine_cluster_node_status(response['cluster_nodes'])
            # Now our status should be STATUS_GOOD or STATUS_DEGRADED
            # There is one more check we need to do and that is if the cluster_info did not return ok we are in a bad state
            if response['cluster_info']['cluster_state'] != _REDIS_CLUSTER_OK_STATUS:
                response['status'] = STATUS_FAILED
    except Exception as e:
        response['status'] = STATUS_FAILED
        response['exception'] = str(e)
        logger.exception("Failed getting redis status")

    return response


def determine_cluster_node_status(cluster_nodes: Dict) -> str:
    # A general marker that there is at least 1 failed node in the cluster
    has_bad_node = False
    for response_node in cluster_nodes.values():
        if 'fail' in response_node['flags']:
            has_bad_node = True

    if has_bad_node:
        return STATUS_DEGRADED

    return STATUS_GOOD



# Generated by Django 4.2.8 on 2024-04-10 20:53

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('dab_authentication', '0007_remove_authenticator_users_and_more'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='authenticator',
            name='users_unique',
        ),
    ]



from base64 import b64encode
from typing import Any

from django.conf import settings
from django.db import models
from django.utils.translation import gettext_lazy as _
from social_django.models import AbstractUserSocialAuth

from ansible_base.authentication.models import Authenticator
from ansible_base.lib.abstract_models.common import AbstractCommonModel
from ansible_base.lib.utils.encryption import ansible_encryption


def b64_encode_binary_data_in_dict(obj: Any) -> Any:
    if isinstance(obj, list):
        for index in range(0, len(obj)):
            obj[index] = b64_encode_binary_data_in_dict(obj[index])
    elif isinstance(obj, dict):
        for key, value in obj.items():
            obj[key] = b64_encode_binary_data_in_dict(value)
    elif isinstance(obj, bytes):
        return b64encode(obj).decode("utf-8")
    return obj


class AuthenticatorUser(AbstractUserSocialAuth, AbstractCommonModel):
    """
    This appends extra information on the local user model that includes extra data returned by
    the authenticators and links the user to the authenticator that they used to login.
    """

    provider = models.ForeignKey(
        Authenticator,
        to_field='slug',
        on_delete=models.PROTECT,
        related_name="authenticator_providers",
        help_text=_("The provider this user authenticated from."),
    )
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        related_name="authenticator_users",
        on_delete=models.CASCADE,
        help_text=_("The local DB user related to this authenticator user."),
    )
    # TODO: set self.authenticated based on the provider that is passed to this method.
    # the provider should be the name of the Authenticator model instance
    claims = models.JSONField(
        default=dict, null=False, blank=True, help_text=_("The claims for the user as generated by the authenticator maps on their last login.")
    )
    last_login_map_results = models.JSONField(
        default=list, null=False, blank=True, help_text=_("A data structure indicating how the authenticator maps were evaluated for the last login attempt.")
    )
    # This field tracks if a user passed or failed an allow map
    access_allowed = models.BooleanField(
        default=None, null=True, help_text=_("Tracks if this user was allowed access to the system from the authenticator maps.")
    )

    encrypted_fields = ["extra_data"]

    @classmethod
    def create_social_auth(cls, user, uid, slug):
        provider = Authenticator.objects.get(slug=slug)
        return super().create_social_auth(user, uid, provider)

    class Meta:
        """Meta data"""

        unique_together = ("provider", "uid")

    def save(self, *args, **kwargs):
        # Some authenticators can put binary field in the extra data so we need to be sure to strip that out.
        self.extra_data = b64_encode_binary_data_in_dict(self.extra_data)

        super().save(*args, **kwargs)

        # After super().save extra_data will be encrypted but we need this to field to not be encrypted during the login process
        self.extra_data = ansible_encryption.decrypt_string(self.extra_data)



# Generated by Django 4.2.8 on 2024-05-09 16:12

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('dab_oauth2_provider', '0002_alter_oauth2refreshtoken_options_and_more'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='oauth2application',
            name='logo_data',
        ),
    ]



# Loads the output from settings_logic into locals for this to be included
# NOTE: This whole module is now deprecated and must be removed soon
# when all consumers are migrated to the new dynamic settings using Dynaconf.

from ansible_base.lib.dynamic_config.settings_logic import get_dab_settings

try:
    ANSIBLE_BASE_OVERRIDABLE_SETTINGS  # noqa: F821
except NameError:
    ANSIBLE_BASE_OVERRIDABLE_SETTINGS = [
        'INSTALLED_APPS',
        'REST_FRAMEWORK',
        'AUTHENTICATION_BACKENDS',
        'SPECTACULAR_SETTINGS',
        'MIDDLEWARE',
        'OAUTH2_PROVIDER',
        'CACHES',
        'TEMPLATES',
    ]


# This is mostly to be informative to the client app
# and this would declare what settings are modified
ANSIBLE_BASE_OVERRIDDEN_SETTINGS = []


# Use temporary local variable so we can avoid setting it if not present
local_authentication_backends = []


# Any settings that will be _modified_ will be mentioned here
try:
    INSTALLED_APPS  # noqa: F821
except NameError:
    INSTALLED_APPS = []

try:
    REST_FRAMEWORK  # noqa: F821
except NameError:
    REST_FRAMEWORK = {}

try:
    local_authentication_backends = AUTHENTICATION_BACKENDS  # noqa: F821
except NameError:
    pass

try:
    SPECTACULAR_SETTINGS  # noqa: F821
except NameError:
    SPECTACULAR_SETTINGS = {}

try:
    MIDDLEWARE  # noqa: F821
except NameError:
    MIDDLEWARE = []

try:
    OAUTH2_PROVIDER  # noqa: F821
except NameError:
    OAUTH2_PROVIDER = {}

try:
    TEMPLATES  # noqa: F821
except NameError:
    TEMPLATES = []

for key, value in get_dab_settings(
    installed_apps=INSTALLED_APPS,
    rest_framework=REST_FRAMEWORK,
    spectacular_settings=SPECTACULAR_SETTINGS,
    authentication_backends=local_authentication_backends,
    middleware=MIDDLEWARE,
    oauth2_provider=OAUTH2_PROVIDER,
    caches=CACHES if 'CACHES' in locals() else None,  # noqa: F821
    templates=TEMPLATES,
).items():
    if key in ANSIBLE_BASE_OVERRIDABLE_SETTINGS:
        ANSIBLE_BASE_OVERRIDDEN_SETTINGS.append(key)
        locals()[key] = value
    elif key not in locals():
        locals()[key] = value


del get_dab_settings
del local_authentication_backends



# Generated by Django 4.2.8 on 2024-03-29 19:51

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('dab_resource_registry', '0003_alter_resource_object_id'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='resourcetype',
            name='migrated',
        ),
    ]



from rest_framework.permissions import SAFE_METHODS, BasePermission

from ansible_base.lib.utils.models import is_system_user
from ansible_base.lib.utils.settings import get_setting

oauth2_provider_installed = "ansible_base.oauth2_provider" in get_setting("INSTALLED_APPS", [])


def try_add_oauth2_scope_permission(permission_classes: list):
    """
    Attach OAuth2ScopePermission to the provided permission_classes list

    :param permission_classes: list of rest_framework permissions
    :return: A list of permission_classes, including OAuth2ScopePermission
        if ansible_base.oauth2_provider is installed; otherwise the same
        permission_classes list supplied to the function
    """
    if oauth2_provider_installed:
        from ansible_base.oauth2_provider.permissions import OAuth2ScopePermission

        return [OAuth2ScopePermission] + permission_classes
    return permission_classes


def check_service_token_auth(request, view):
    if is_system_user(request.user) and getattr(view, 'allow_service_token', False) is True and request.auth == 'ServiceTokenAuthentication':
        return True
    return False


class IsSuperuser(BasePermission):
    """
    Allows access only to superusers.
    """

    def has_permission(self, request, view):
        if check_service_token_auth(request, view):
            return True

        return bool(request.user and request.user.is_authenticated and request.user.is_superuser)


class IsSuperuserOrAuditor(BasePermission):
    """
    Allows write access only to system admin users.
    Allows read access only to system auditor users.
    """

    def has_permission(self, request, view):
        if check_service_token_auth(request, view):
            return True
        if not (request.user and request.user.is_authenticated):
            return False
        if request.user.is_superuser:
            return True
        if request.method in SAFE_METHODS:
            return getattr(request.user, 'is_platform_auditor', False)
        return False



import logging
from copy import deepcopy
from typing import Optional

from rest_framework import serializers

from ansible_base.activitystream.models import Entry
from ansible_base.lib.abstract_models.common import CreatableModel, get_url_for_object
from ansible_base.lib.serializers.common import ImmutableCommonModelSerializer

logger = logging.getLogger('ansible_base.activitystream.serializers')


class EntrySerializer(ImmutableCommonModelSerializer):
    class Meta:
        model = Entry
        fields = ImmutableCommonModelSerializer.Meta.fields + [
            'operation',
            'changes',
            'content_type',
            'object_id',
            'related_content_type',
            'related_object_id',
            'content_type_model',
            'related_content_type_model',
        ]

    content_type_model = serializers.SerializerMethodField()
    related_content_type_model = serializers.SerializerMethodField()
    changes = serializers.SerializerMethodField()

    def get_content_type_model(self, obj) -> Optional[str]:
        if obj.content_type:
            return obj.content_type.model

    def get_related_content_type_model(self, obj) -> Optional[str]:
        if obj.related_content_type:
            return obj.related_content_type.model

    def _field_value_to_python(self, entry, field_name, value):
        model = entry.content_type.model_class()
        if model is None:
            # If the model was deleted, we lose the ability to convert the
            # field back to the correct type, because we don't know what the
            # field was. We'll just return the value as is and keep it as a
            # string. This is kind of a gross edge case.
            return value
        field = model._meta.get_field(field_name)
        return field.to_python(value)

    def get_changes(self, obj) -> Optional[dict[str, dict]]:
        """
        We store strings, we have to convert them back to the correct type.
        Related associations and disassociations will show a null for changes.
        """
        if not obj.changes:
            return None

        changes = deepcopy(obj.changes)
        # We'll have 'added_fields', 'removed_fields', 'changed_fields'. The first two
        # are simple k-v pairs, the last is a k-v pair where the value is [old, new].
        for field_name, value in obj.changes['added_fields'].items():
            changes['added_fields'][field_name] = self._field_value_to_python(obj, field_name, value)
        for field_name, value in obj.changes['removed_fields'].items():
            changes['removed_fields'][field_name] = self._field_value_to_python(obj, field_name, value)
        for field_name, value in obj.changes['changed_fields'].items():
            changes['changed_fields'][field_name] = [
                self._field_value_to_python(obj, field_name, value[0]),
                self._field_value_to_python(obj, field_name, value[1]),
            ]
        return changes

    def _get_summary_fields(self, obj) -> dict[str, dict]:
        summary_fields = super()._get_summary_fields(obj)
        if self.is_list_view:
            return summary_fields

        try:
            if obj.content_object is not None and hasattr(obj.content_object, 'summary_fields'):
                summary_fields['content_object'] = obj.content_object.summary_fields()
        except AttributeError:  # Likely the model was deleted
            pass

        try:
            if obj.related_content_object is not None and hasattr(obj.related_content_object, 'summary_fields'):
                summary_fields['related_content_object'] = obj.related_content_object.summary_fields()
        except AttributeError:  # Likely the model was deleted
            pass

        if obj.changes is None:
            return summary_fields

        changed_fk_fields = obj.changed_fk_fields

        for field_name, (related_model, pk) in changed_fk_fields.items():
            if related_object := related_model.objects.filter(pk=pk).first():
                if hasattr(related_object, 'summary_fields'):
                    summary_fields[f"changes.{field_name}"] = related_object.summary_fields()

        return summary_fields

    def _get_related(self, obj) -> dict[str, str]:
        fields = super()._get_related(obj)

        if self.is_list_view:
            return fields

        # content_object
        try:
            if obj.content_object is not None:
                fields['content_object'] = get_url_for_object(obj.content_object)
        except AttributeError:  # Likely the model was deleted
            pass

        # related_content_object
        try:
            if obj.related_content_object is not None:
                fields['related_content_object'] = get_url_for_object(obj.related_content_object)
        except AttributeError:  # Likely the model was deleted
            pass

        for field_name, (related_model, pk) in obj.changed_fk_fields.items():
            if related_object := related_model.objects.filter(pk=pk).first():
                # If the related object inherits CreatableModel, we can check and make sure it's
                # older than the activity stream entry. If it's not, then we don't want to link to it.
                if isinstance(related_object, CreatableModel) and related_object.created > obj.created:
                    model = obj.content_type.model_class()
                    logger.warning(
                        f"Refusing to relate {related_object} to activity stream entry for {model} {obj.object_id} because it was created after the entry"
                    )
                    continue

                if related_url := get_url_for_object(related_object, pk=pk):
                    fields[f"changes.{field_name}"] = related_url

        return fields



from django.apps import apps
from django.core.exceptions import ObjectDoesNotExist
from django.db import transaction
from django.db.utils import IntegrityError
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _
from rest_framework import serializers
from rest_framework.exceptions import PermissionDenied
from rest_framework.fields import flatten_choices_dict, to_choices_dict
from rest_framework.serializers import ValidationError

from ansible_base.lib.abstract_models.common import get_url_for_object
from ansible_base.lib.serializers.common import CommonModelSerializer, ImmutableCommonModelSerializer
from ansible_base.rbac.models import RoleDefinition, RoleTeamAssignment, RoleUserAssignment
from ansible_base.rbac.permission_registry import permission_registry  # careful for circular imports
from ansible_base.rbac.policies import check_content_obj_permission, visible_users
from ansible_base.rbac.validators import check_locally_managed, validate_permissions_for_model


class ChoiceLikeMixin(serializers.ChoiceField):
    """
    This uses a ForeignKey to populate the choices of a choice field.
    This also manages some string manipulation, right now, adding the local service name.
    """

    default_error_messages = serializers.PrimaryKeyRelatedField.default_error_messages

    def get_dynamic_choices(self):
        raise NotImplementedError

    def get_dynamic_object(self, data):
        raise NotImplementedError

    def to_representation(self, value):
        raise NotImplementedError

    def __init__(self, **kwargs):
        # Workaround so that the parent class does not resolve the choices right away
        self.html_cutoff = kwargs.pop('html_cutoff', self.html_cutoff)
        self.html_cutoff_text = kwargs.pop('html_cutoff_text', self.html_cutoff_text)

        self.allow_blank = kwargs.pop('allow_blank', False)
        super(serializers.ChoiceField, self).__init__(**kwargs)

    def _initialize_choices(self):
        choices = self.get_dynamic_choices()
        self._grouped_choices = to_choices_dict(choices)
        self._choices = flatten_choices_dict(self._grouped_choices)
        self.choice_strings_to_values = {str(k): k for k in self._choices}

    @cached_property
    def grouped_choices(self):
        self._initialize_choices()
        return self._grouped_choices

    @cached_property
    def choices(self):
        self._initialize_choices()
        return self._choices

    def to_internal_value(self, data):
        try:
            return self.get_dynamic_object(data)
        except ObjectDoesNotExist:
            self.fail('does_not_exist', pk_value=data)
        except (TypeError, ValueError):
            self.fail('incorrect_type', data_type=type(data).__name__)


class ContentTypeField(ChoiceLikeMixin):
    def __init__(self, **kwargs):
        kwargs['help_text'] = _('The type of resource this applies to.')
        super().__init__(**kwargs)

    def get_resource_type_name(self, cls) -> str:
        return f"{permission_registry.get_resource_prefix(cls)}.{cls._meta.model_name}"

    def get_dynamic_choices(self):
        return list(sorted((self.get_resource_type_name(cls), cls._meta.verbose_name.title()) for cls in permission_registry.all_registered_models))

    def get_dynamic_object(self, data):
        model = data.rsplit('.')[-1]
        cls = permission_registry.get_model_by_name(model)
        if cls is None:
            return permission_registry.content_type_model.objects.none().get()  # raises correct DoesNotExist
        return permission_registry.content_type_model.objects.get_for_model(cls)

    def to_representation(self, value):
        if isinstance(value, str):
            return value  # slight hack to work to AWX schema tests
        return self.get_resource_type_name(value.model_class())


class PermissionField(ChoiceLikeMixin):
    @property
    def service_prefix(self):
        if registry := permission_registry.get_resource_registry():
            return registry.api_config.service_type
        return 'local'

    def get_dynamic_choices(self):
        perms = []
        for cls in permission_registry.all_registered_models:
            cls_name = cls._meta.model_name
            for action in cls._meta.default_permissions:
                perms.append(f'{permission_registry.get_resource_prefix(cls)}.{action}_{cls_name}')
            for perm_name, description in cls._meta.permissions:
                perms.append(f'{permission_registry.get_resource_prefix(cls)}.{perm_name}')
        return list(sorted(perms))

    def get_dynamic_object(self, data):
        codename = data.rsplit('.')[-1]
        return permission_registry.permission_qs.get(codename=codename)

    def to_representation(self, value):
        if isinstance(value, str):
            return value  # slight hack to work to AWX schema tests
        ct = permission_registry.content_type_model.objects.get_for_id(value.content_type_id)  # optimization
        return f'{permission_registry.get_resource_prefix(ct.model_class())}.{value.codename}'


class ManyRelatedListField(serializers.ListField):
    def to_representation(self, data):
        "Adds the .all() to treat the value as a queryset"
        return [self.child.to_representation(item) if item is not None else None for item in data.all()]


class RoleDefinitionSerializer(CommonModelSerializer):
    # Relational versions - we may switch to these if custom permission and type models are exposed but out of scope here
    # permissions = serializers.SlugRelatedField(many=True, slug_field='codename', queryset=DABPermission.objects.all())
    # content_type = ContentTypeField(slug_field='model', queryset=permission_registry.content_type_model.objects.all(), allow_null=True, default=None)
    permissions = ManyRelatedListField(child=PermissionField())
    content_type = ContentTypeField(allow_null=True, default=None)

    class Meta:
        model = RoleDefinition
        read_only_fields = ('id', 'summary_fields')
        fields = '__all__'

    def validate(self, validated_data):
        # Obtain the resultant new values
        if 'permissions' in validated_data:
            permissions = validated_data['permissions']
        else:
            permissions = list(self.instance.permissions.all())
        if 'content_type' in validated_data:
            content_type = validated_data['content_type']
        else:
            content_type = self.instance.content_type
        validate_permissions_for_model(permissions, content_type)
        if getattr(self, 'instance', None):
            check_locally_managed(self.instance)
        return super().validate(validated_data)


class RoleDefinitionDetailSerializer(RoleDefinitionSerializer):
    content_type = ContentTypeField(read_only=True)


class BaseAssignmentSerializer(CommonModelSerializer):
    content_type = ContentTypeField(read_only=True)
    object_ansible_id = serializers.UUIDField(
        required=False,
        help_text=_('The resource id of the object this role applies to. An alternative to the object_id field.'),
        allow_null=True,  # for ease of use of the browseable API
    )

    def __init__(self, *args, **kwargs):
        """
        We want to allow ansible_id override of user and team fields
        but want to keep the non-null database constraint, so actor field is marked required=True here
        """
        super().__init__(*args, **kwargs)
        request = self.context.get('request')
        if request:
            qs = self.get_actor_queryset(request.user)
        else:
            qs = self.Meta.model._meta.get_field(self.actor_field).model.objects.all()
        self.fields[self.actor_field] = serializers.PrimaryKeyRelatedField(queryset=qs, required=False)

    def raise_id_fields_error(self, field1, field2):
        msg = _('Provide exactly one of %(actor_field)s or %(actor_field)s_ansible_id') % {'actor_field': self.actor_field}
        raise ValidationError({self.actor_field: msg, f'{self.actor_field}_ansible_id': msg})

    def get_by_ansible_id(self, ansible_id, requesting_user, for_field):
        try:
            resource_cls = apps.get_model('dab_resource_registry', 'Resource')
        except LookupError:
            raise ValidationError({for_field: _('Django-ansible-base resource registry must be installed to use ansible_id fields')})

        try:
            resource = resource_cls.objects.get(ansible_id=ansible_id)
            # Ensure that the request user has permission to view provided data
            obj = resource.content_object
            if obj._meta.model_name == 'user':
                if not visible_users(requesting_user).filter(pk=obj.pk).exists():
                    raise ObjectDoesNotExist
            elif not requesting_user.has_obj_perm(obj, 'view'):
                raise ObjectDoesNotExist
        except ObjectDoesNotExist:
            msg = serializers.PrimaryKeyRelatedField.default_error_messages['does_not_exist']
            raise ValidationError({for_field: msg.format(pk_value=ansible_id)})
        return resource.content_object

    def get_actor_from_data(self, validated_data, requesting_user):
        actor_aid_field = f'{self.actor_field}_ansible_id'
        if validated_data.get(self.actor_field) and validated_data.get(actor_aid_field):
            self.raise_id_fields_error(self.actor_field, actor_aid_field)
        elif validated_data.get(self.actor_field):
            actor = validated_data[self.actor_field]
        elif ansible_id := validated_data.get(actor_aid_field):
            actor = self.get_by_ansible_id(ansible_id, requesting_user, for_field=actor_aid_field)
        else:
            self.raise_id_fields_error(self.actor_field, f'{self.actor_field}_ansible_id')
        return actor

    def get_object_from_data(self, validated_data, role_definition, requesting_user):
        obj = None
        if validated_data.get('object_id') and validated_data.get('object_ansible_id'):
            self.raise_id_fields_error('object_id', 'object_ansible_id')
        elif validated_data.get('object_id'):
            if not role_definition.content_type:
                raise ValidationError({'object_id': _('System role does not allow for object assignment')})
            model = role_definition.content_type.model_class()
            try:
                obj = serializers.PrimaryKeyRelatedField(queryset=model.access_qs(requesting_user)).to_internal_value(validated_data['object_id'])
            except ValidationError as exc:
                raise ValidationError({'object_id': exc.detail})
            except AttributeError:
                if not permission_registry.is_registered(model):
                    raise ValidationError({'role_definition': 'Given role definition is for a model not registered in the permissions system'})
                raise  # in this case no idea what went wrong
        elif validated_data.get('object_ansible_id'):
            obj = self.get_by_ansible_id(validated_data.get('object_ansible_id'), requesting_user, for_field='object_ansible_id')
            if permission_registry.content_type_model.objects.get_for_model(obj) != role_definition.content_type:
                raise ValidationError(
                    {
                        'object_ansible_id': _('Object type of %(model_name)s does not match role type of %(role_definition)s')
                        % {'model_name': obj._meta.model_name, 'role_definition': role_definition.content_type.model}
                    }
                )
        return obj

    def create(self, validated_data):
        rd = validated_data['role_definition']
        requesting_user = self.context['view'].request.user

        # Resolve actor - team or user
        actor = self.get_actor_from_data(validated_data, requesting_user)

        # Resolve object
        obj = self.get_object_from_data(validated_data, rd, requesting_user)

        # model-level callback to further validate the assignment
        # can be optionally implemented by the model
        # the callback should raise DRF exceptions directly if
        # necessary
        if getattr(obj, 'validate_role_assignment', None):
            obj.validate_role_assignment(actor, rd, requesting_user=requesting_user)

        # Return a 400 if the role is not managed locally
        check_locally_managed(rd)

        if rd.content_type:
            # Object role assignment
            if not obj:
                raise ValidationError({'object_id': _('Object must be specified for this role assignment')})

            check_content_obj_permission(requesting_user, obj)

            try:
                with transaction.atomic():
                    assignment = rd.give_permission(actor, obj)
            except IntegrityError:
                assignment = self.Meta.model.objects.get(role_definition=rd, object_id=obj.pk, **{self.actor_field: actor})
        else:
            # Global role assignment, only allowed by superuser
            if not requesting_user.is_superuser:
                raise PermissionDenied

            with transaction.atomic():
                assignment = rd.give_global_permission(actor)

        return assignment

    def _get_related(self, obj) -> dict[str, str]:
        related = super()._get_related(obj)
        content_obj = obj.content_object
        if content_obj:
            if related_url := get_url_for_object(content_obj):
                related['content_object'] = related_url
        return related

    def _get_summary_fields(self, obj) -> dict[str, dict]:
        summary_fields = super()._get_summary_fields(obj)
        content_obj = obj.content_object
        if content_obj and hasattr(content_obj, 'summary_fields'):
            summary_fields['content_object'] = content_obj.summary_fields()
        return summary_fields


ASSIGNMENT_FIELDS = ImmutableCommonModelSerializer.Meta.fields + ['content_type', 'object_id', 'object_ansible_id', 'role_definition']


class RoleUserAssignmentSerializer(BaseAssignmentSerializer):
    actor_field = 'user'
    user_ansible_id = serializers.UUIDField(
        required=False,
        help_text=_('The resource ID of the user who will receive permissions from this assignment. An alternative to user field.'),
        allow_null=True,  # for ease of use of the browseable API
    )

    class Meta:
        model = RoleUserAssignment
        fields = ASSIGNMENT_FIELDS + ['user', 'user_ansible_id']

    def get_actor_queryset(self, requesting_user):
        return visible_users(requesting_user)


class RoleTeamAssignmentSerializer(BaseAssignmentSerializer):
    actor_field = 'team'
    team_ansible_id = serializers.UUIDField(
        required=False,
        help_text=_('The resource ID of the team who will receive permissions from this assignment. An alternative to team field.'),
    )

    class Meta:
        model = RoleTeamAssignment
        fields = ASSIGNMENT_FIELDS + ['team', 'team_ansible_id']

    def get_actor_queryset(self, requesting_user):
        return permission_registry.team_model.access_qs(requesting_user)


class RoleMetadataSerializer(serializers.Serializer):
    allowed_permissions = serializers.DictField(help_text=_('A List of permissions allowed for a role definition, given its content type.'))



import importlib
import logging
from types import ModuleType

from django.conf import settings
from django.contrib.auth import get_user_model
from django.http import Http404

from ansible_base.authentication.models import Authenticator
from ansible_base.lib.utils.views.permissions import IsSuperuserOrAuditor, try_add_oauth2_scope_permission

logger = logging.getLogger('ansible_base.authentication.views.authenticator_users')


def get_authenticator_user_view():
    try:
        user_viewset_name = settings.ANSIBLE_BASE_USER_VIEWSET
        module_name, _junk, class_name = user_viewset_name.rpartition('.')
        module = importlib.import_module(module_name, package=class_name)
        user_viewset_view = getattr(module, class_name)
        if isinstance(user_viewset_view, ModuleType):
            raise ModuleNotFoundError()

        class AuthenticatorPluginRelatedUsersView(user_viewset_view):
            permission_classes = try_add_oauth2_scope_permission([IsSuperuserOrAuditor])

            def get_queryset(self, **kwargs):
                # during unit testing we get the pk from kwargs
                authenticator_id = kwargs.get('pk', None)
                if hasattr(self, 'kwargs'):
                    # But at runtime self has kwargs attached and no kwargs associated
                    authenticator_id = self.kwargs.get('pk', None)
                # if we didn't get an ID for some reason we will just return None
                if authenticator_id is None or not Authenticator.objects.filter(pk=authenticator_id).exists():
                    raise Http404()
                authenticator_users = get_user_model().objects.filter(authenticator_users__provider__id=authenticator_id)
                return authenticator_users

        return AuthenticatorPluginRelatedUsersView
    except ModuleNotFoundError:
        logger.error("ANSIBLE_BASE_USER_VIEWSET was not an APIView")
    except AttributeError:
        logger.debug("ANSIBLE_BASE_USER_VIEWSET was not specified")
    except Http404:
        logger.error("Authenticator not available")

    return None



import contextlib
import importlib
import logging
import re
from enum import Enum, auto
from typing import Optional, Union

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import AbstractUser
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ObjectDoesNotExist
from django.db import IntegrityError, models
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _
from rest_framework.serializers import DateTimeField

from ansible_base.authentication.models import Authenticator, AuthenticatorMap, AuthenticatorUser
from ansible_base.lib.abstract_models import AbstractOrganization, AbstractTeam, CommonModel
from ansible_base.lib.utils.auth import get_organization_model, get_team_model
from ansible_base.lib.utils.string import is_empty

from .trigger_definition import TRIGGER_DEFINITION

logger = logging.getLogger('ansible_base.authentication.utils.claims')
Organization = get_organization_model()
Team = get_team_model()
User = get_user_model()


class TriggerResult(Enum):
    ALLOW = auto()
    DENY = auto()
    SKIP = auto()


def create_claims(authenticator: Authenticator, username: str, attrs: dict, groups: list[str]) -> dict:
    """
    Given an authenticator and a username, attrs and groups determine what the user has access to
    """

    # Assume we are not going to change our flags
    is_superuser = None

    rbac_role_mapping = {'system': {'roles': {}}, 'organizations': {}}
    # Assume we start with no mappings
    org_team_mapping = {}
    # Assume we are not members of any orgs (direct members)
    organization_membership = {}
    # Start with an empty rule responses
    rule_responses = []
    # Assume we will have access
    access_allowed = True
    logger.info(f"Creating mapping for user {username} through authenticator {authenticator.name}")
    logger.debug(f"{username}'s groups: {groups}")
    logger.debug(f"{username}'s attrs: {attrs}")

    # load the maps
    logger.debug(f"Authenticator ID: {authenticator.id}")
    maps = AuthenticatorMap.objects.order_by("order")
    logger.debug(maps)
    maps = AuthenticatorMap.objects.filter(authenticator=authenticator.id).order_by("order")
    logger.debug("==============================================================")
    logger.debug(maps)

    for auth_map in maps:
        logger.debug(auth_map)
        logger.debug("++++")
        has_permission = None
        trigger_result = TriggerResult.SKIP
        allowed_keys = TRIGGER_DEFINITION.keys()
        invalid_keys = set(auth_map.triggers.keys()) - set(allowed_keys)

        if auth_map.enabled is False:
            logger.info(f"Skipping AuthenticatorMap {auth_map.id} because it is disabled")
            rule_responses.append({auth_map.id: 'skipped', 'enabled': auth_map.enabled})
            continue

        if invalid_keys:
            logger.warning(f"In AuthenticatorMap {auth_map.id} the following trigger keys are invalid: {', '.join(invalid_keys)}, rule will be ignored")
            rule_responses.append({auth_map.id: 'invalid', 'enabled': auth_map.enabled})
            continue

        for trigger_type, trigger in auth_map.triggers.items():
            if trigger_type == 'groups':
                trigger_result = process_groups(trigger, groups, authenticator.pk)
            elif trigger_type == 'attributes':
                trigger_result = process_user_attributes(trigger, attrs, authenticator.pk)
            elif trigger_type == 'always':
                trigger_result = TriggerResult.ALLOW
            elif trigger_type == 'never':
                trigger_result = TriggerResult.DENY

        # If the trigger result is SKIP, auth map is not defined for this user.
        # Together with "revoke" flag => change permission to DENY
        if auth_map.revoke and trigger_result is TriggerResult.SKIP:
            trigger_result = TriggerResult.DENY

        # If the trigger result is still SKIP, this auth map is not applicable to this user => no action needed
        if trigger_result is TriggerResult.SKIP:
            rule_responses.append({auth_map.id: 'skipped', 'enabled': auth_map.enabled})
            continue

        if trigger_result is TriggerResult.ALLOW:
            has_permission = True
        elif trigger_result is TriggerResult.DENY:
            has_permission = False

        rule_responses.append({auth_map.id: has_permission, 'enabled': auth_map.enabled})

        if auth_map.map_type == 'allow' and not has_permission:
            # If any rule does not allow we don't want to return this to true
            access_allowed = False
        elif auth_map.map_type == 'is_superuser':
            is_superuser = has_permission
        elif auth_map.map_type in ['team', 'role'] and not is_empty(auth_map.organization) and not is_empty(auth_map.team) and not is_empty(auth_map.role):
            if auth_map.organization not in org_team_mapping:
                org_team_mapping[auth_map.organization] = {}
            org_team_mapping[auth_map.organization][auth_map.team] = has_permission
            _add_rbac_role_mapping(has_permission, rbac_role_mapping, auth_map.role, auth_map.organization, auth_map.team)
        elif auth_map.map_type in ['organization', 'role'] and not is_empty(auth_map.organization) and not is_empty(auth_map.role):
            organization_membership[auth_map.organization] = has_permission
            _add_rbac_role_mapping(has_permission, rbac_role_mapping, auth_map.role, auth_map.organization)
        elif auth_map.map_type == 'role' and not is_empty(auth_map.role) and is_empty(auth_map.organization) and is_empty(auth_map.team):
            _add_rbac_role_mapping(has_permission, rbac_role_mapping, auth_map.role)
        else:
            logger.error(f"Map type {auth_map.map_type} of rule {auth_map.name} does not know how to be processed")

    return {
        "access_allowed": access_allowed,
        "is_superuser": is_superuser,
        "claims": {
            "team_membership": org_team_mapping,
            "organization_membership": organization_membership,
            "rbac_roles": rbac_role_mapping,
        },
        "last_login_map_results": rule_responses,
    }


def _add_rbac_role_mapping(has_permission, role_mapping, role, organization=None, team=None):
    """
    Example of RBAC roles mapping dict:
    {
      'system': {'roles': {'System Auditor': true}},
      'organizations': {
        'Organization 1': {
            'roles': {'Organization Member': true, 'Organization Admin': false},
            'teams': {}
         },
        'Organization 2': {
            'roles': {'Organization Admin': true},
            'teams': {
                'Team 1': {
                    'roles': {'Team Member': true},
                },
                'Team 2': {
                    'roles': {'Team Admin': false},
                }
            }
        }
    """
    # System role
    if organization is None and team is None:
        role_mapping['system']['roles'][role] = has_permission
    else:
        if organization not in role_mapping['organizations']:
            role_mapping['organizations'][organization] = {'roles': {}, 'teams': {}}
        # Organization role
        if organization and not team:
            role_mapping['organizations'][organization]['roles'][role] = has_permission
        # Team role
        elif organization and team:
            if team not in role_mapping['organizations'][organization]['teams']:
                role_mapping['organizations'][organization]['teams'][team] = {'roles': {}}
            role_mapping['organizations'][organization]['teams'][team]['roles'][role] = has_permission
        else:
            logger.warning(f"Role mapping is not possible, organization for team '{team}' is missing")


def process_groups(trigger_condition: dict, groups: list, authenticator_id: int) -> TriggerResult:
    """
    Looks at a maps trigger for a group and users groups and determines if the trigger is defined for this user.
    """

    invalid_conditions = set(trigger_condition.keys()) - set(TRIGGER_DEFINITION['groups']['keys'].keys())
    if invalid_conditions:
        logger.warning(f"The conditions {', '.join(invalid_conditions)} for groups in mapping {authenticator_id} are invalid and won't be processed")

    set_of_user_groups = set(groups)

    if "has_or" in trigger_condition:
        if set_of_user_groups.intersection(set(trigger_condition["has_or"])):
            return TriggerResult.ALLOW

    elif "has_and" in trigger_condition:
        if set(trigger_condition["has_and"]).issubset(set_of_user_groups):
            return TriggerResult.ALLOW

    elif "has_not" in trigger_condition:
        if not set(trigger_condition["has_not"]).intersection(set_of_user_groups):
            return TriggerResult.ALLOW

    return TriggerResult.SKIP


def has_access_with_join(current_access: Optional[bool], new_access: bool, condition: str = 'or') -> Optional[bool]:
    """
    Handle join of authenticator_maps
    """
    if current_access is None:
        return new_access

    if condition == 'or':
        return current_access or new_access

    if condition == 'and':
        return current_access and new_access


def process_user_attributes(trigger_condition: dict, attributes: dict, authenticator_id: int) -> TriggerResult:
    """
    Looks at a maps trigger for an attribute and the users attributes and determines if the trigger is defined for this user.
    """
    has_access = None
    join_condition = trigger_condition.get('join_condition', 'or')
    if join_condition not in TRIGGER_DEFINITION['attributes']['keys']['join_condition']['choices']:
        logger.warning("Trigger join_condition {join_condition} on authenticator map {authenticator_id} is invalid and will be set to 'or'")
        join_condition = 'or'

    for attribute in trigger_condition.keys():
        if has_access and join_condition == 'or':
            # If we are an or condition and we already have a positive we can break out and return
            break
        elif has_access is False and join_condition == 'and':
            # If we are an and and already have a False we can give up
            break

        # We can skip the join_condition since we already processed that.
        if attribute == 'join_condition':
            continue

        # Warn if there are any invalid conditions, we are just going to ignore them
        invalid_conditions = set(trigger_condition[attribute].keys()) - set(TRIGGER_DEFINITION['attributes']['keys']['*']['keys'].keys())
        if invalid_conditions:
            logger.warning(
                f"The conditions {', '.join(invalid_conditions)} for attribute {attribute} "
                "in authenticator map {authenticator_id} are invalid and won't be processed"
            )

        # The attribute is an empty dict we just need to see if the user has the attribute or not
        if trigger_condition[attribute] == {}:
            has_access = has_access_with_join(has_access, attribute in attributes, join_condition)
            continue

        user_value = attributes.get(attribute, None)
        # If the user does not contain the attribute then we can't check any further, don't set has_access and just continue
        if user_value is None:
            continue

        if type(user_value) is not list:
            # If the value is a string then convert it to a list
            user_value = [user_value]

        for a_user_value in user_value:
            # We are going to do mostly string comparisons, so convert the attribute to a
            #  string just in case it came back as an int or something funky
            a_user_value = f"{a_user_value}"

            # Check for any of the valid conditions
            if "equals" in trigger_condition[attribute]:
                has_access = has_access_with_join(has_access, a_user_value == trigger_condition[attribute]["equals"], join_condition)

            elif "matches" in trigger_condition[attribute]:
                has_access = has_access_with_join(
                    has_access, re.match(trigger_condition[attribute]["matches"], a_user_value, re.IGNORECASE) is not None, join_condition
                )

            elif "contains" in trigger_condition[attribute]:
                has_access = has_access_with_join(has_access, trigger_condition[attribute]['contains'] in a_user_value, join_condition)

            elif "ends_with" in trigger_condition[attribute]:
                has_access = has_access_with_join(has_access, a_user_value.endswith(trigger_condition[attribute]['ends_with']), join_condition)

            elif "in" in trigger_condition[attribute]:
                has_access = has_access_with_join(has_access, a_user_value in trigger_condition[attribute]['in'], join_condition)

    return TriggerResult.ALLOW if has_access else TriggerResult.SKIP


def update_user_claims(user: Optional[AbstractUser], database_authenticator: Authenticator, groups: list[str]) -> Optional[AbstractUser]:
    """
    This method takes a user, an authenticator and a list of the users associated groups.
    It will look up the AuthenticatorUser (it must exist already) and update that User and their permissions in the system.
    """
    if not user:
        return None

    authenticator_user = user.authenticator_users.filter(provider=database_authenticator).first()
    # update the auth_time field to align with the general format used for other authenticators
    authenticator_user.extra_data = {**authenticator_user.extra_data, "auth_time": DateTimeField().to_representation(now())}
    authenticator_user.save(update_fields=["extra_data"])

    results = create_claims(database_authenticator, user.username, authenticator_user.extra_data, groups)

    needs_save = False

    for attribute, attr_value in results.items():
        if attr_value is None:
            continue
        logger.debug(f"{attribute}: {attr_value}")
        if hasattr(user, attribute):
            object = user
        elif hasattr(authenticator_user, attribute):
            object = authenticator_user
        else:
            logger.error(f"Neither user nor authenticator user has attribute {attribute}")
            continue

        if getattr(object, attribute, None) != attr_value:
            logger.debug(f"Setting new attribute {attribute} for {user.username}")
            setattr(object, attribute, attr_value)
            needs_save = True

    if needs_save:
        authenticator_user.save()
        user.save()
    else:
        # If we don't have to save because of a change we at least need to save the extra data with the login timestamp
        authenticator_user.save(update_fields=["extra_data"])

    if results['access_allowed'] is not True:
        logger.warning(f"User {user.username} failed an allow map and was denied access")
        return None

    # Make the orgs and the teams as necessary ...
    if database_authenticator.create_objects:
        create_organizations_and_teams(results)

    if reconcile_user_class := load_reconcile_user_class():
        try:
            # We have allowed access, so now we need to make the user within the system
            reconcile_user_class.reconcile_user_claims(user, authenticator_user)
        except Exception as e:
            logger.exception("Failed to reconcile user claims: %s", e)
    return user


# TODO(cutwater): Implement a generic version of this function and move it to lib/utils.
def load_reconcile_user_class():
    module_path = getattr(settings, 'ANSIBLE_BASE_AUTHENTICATOR_RECONCILE_MODULE', 'ansible_base.authentication.utils.claims')
    try:
        module = importlib.import_module(module_path)
    except ImportError:
        logger.warning("Failed to load module '%s'.", module_path)
        return None

    try:
        return getattr(module, 'ReconcileUser')
    except AttributeError:
        logger.warning("Failed to load ReconcileUser class in module '%s'.", module_path)
        return None


def create_organizations_and_teams(results) -> None:
    """
    Use the results data from 'create_claims' to make the Organization
    and Team objects necessary for the user if they do not exist.
    """
    # a flat list of relevant org names
    all_orgs = set()
    # a flat list of relevant org:team names
    team_orgs = set()
    # a structure for caching org+team,member info
    membership_map = {}

    # fill in the top level org membership data ...
    for org_name, is_member in results['claims']['organization_membership'].items():
        if not is_member:
            continue
        all_orgs.add(org_name)
        membership_map[org_name] = {'id': None, 'teams': []}

    # fill in the team membership data ...
    for org_name, teams in results['claims']['team_membership'].items():
        for team_name, is_member in teams.items():
            if not is_member:
                continue
            all_orgs.add(org_name)
            team_orgs.add(org_name)
            if org_name not in membership_map:
                membership_map[org_name] = {'id': None, 'teams': []}
            membership_map[org_name]['teams'].append(team_name)

    # Create organizations
    existing_orgs = dict(Organization.objects.filter(name__in=all_orgs).values_list("name", "id"))
    for org_name in all_orgs:
        org_id = existing_orgs.get(org_name)
        if org_id is None:
            try:
                org_id = Organization.objects.create(name=org_name).id
            except IntegrityError:
                org_id = Organization.objects.filter(name=org_name).values_list("id", flat=True).first()
                if org_id is None:
                    raise
        membership_map[org_name]['id'] = org_id

    # Create teams
    # make a map or org id, team name to reduce calls and data sent over the wire
    team_org_ids = [membership_map[org_name]['id'] for org_name in team_orgs]
    existing_teams = set(Team.objects.filter(organization__in=team_org_ids).order_by().values_list('organization', 'name'))
    for org_name, org_data in membership_map.items():
        org_id = org_data['id']
        for team_name in org_data['teams']:
            if (org_id, team_name) not in existing_teams:
                with contextlib.suppress(IntegrityError):
                    Team.objects.create(name=team_name, organization_id=org_id)


# NOTE(cutwater): Current class is sub-optimal, since it loads the data that has been already loaded
#  at the teams and organizations creation. Next step will be combining teams and organizations creation with
#  this class and transforming it into a reconciliation use case class. This implies either
#  removal or update of a pluggable interface.
class ReconcileUser:
    @classmethod
    def reconcile_user_claims(cls, user: AbstractUser, authenticator_user: AuthenticatorUser) -> None:
        logger.info("Reconciling user claims")

        claims = getattr(user, 'claims', authenticator_user.claims)

        if 'ansible_base.rbac' in settings.INSTALLED_APPS:
            cls(claims, user, authenticator_user).manage_permissions()
        else:
            logger.info(_("Skipping user claims with RBAC roles, because RBAC app is not installed"))

    def __init__(self, claims: dict, user: AbstractUser, authenticator_user: AuthenticatorUser):
        """
        :param claims: initialized by method create_claims()
        """
        self.authenticator_user = authenticator_user
        self.claims = claims
        self.permissions_cache = RoleUserAssignmentsCache()
        self.rebuild_user_permissions = self.authenticator_user.provider.remove_users
        self.user = user

    def manage_permissions(self) -> None:
        """
        Processes the user claims (key `rbac_roles`)
        and adds/removes RBAC permissions (a.k.a. role_user_assignments)
        """
        # NOTE(cutwater): Here `prefetch_related` is used to prevent N+1 problem when accessing `content_object`
        #  attribute in `RoleUserAssignmentsCache.cache_existing` method.
        role_assignments = self.user.role_assignments.prefetch_related('content_object').all()
        self.permissions_cache.cache_existing(role_assignments)

        # System roles
        self._compute_system_permissions()

        # Organization roles
        org_info = self._compute_organization_permissions()
        org_teams = self._get_org_teams([org.id for (org, _) in org_info])
        # Team roles
        for org, org_teams_dict in org_info:
            self._compute_team_permissions(org, org_teams_dict, org_teams)

        self.apply_permissions()

    def _compute_system_permissions(self) -> None:
        for role_name, has_permission in self.claims['rbac_roles'].get('system', {}).get('roles', {}).items():
            self.permissions_cache.add_or_remove(role_name, has_permission, organization=None, team=None)

    def _compute_organization_permissions(self) -> list[tuple[AbstractOrganization, dict[str, dict]]]:
        orgs_by_name = self._get_orgs_by_name(self.claims['rbac_roles'].get('organizations', {}).keys())

        org_info = []

        for org_name, org_details in self.claims['rbac_roles'].get('organizations', {}).items():
            if (org := orgs_by_name.get(org_name)) is None:
                logger.error(
                    _("Skipping organization '{organization}', because the organization does not exist but it should already have been created").format(
                        organization=org_name
                    )
                )
                continue

            for role_name, has_permission in org_details['roles'].items():
                self.permissions_cache.add_or_remove(role_name, has_permission, organization=org)

            org_info.append((org, org_details['teams']))
        return org_info

    def _compute_team_permissions(self, org: AbstractOrganization, teams_dict: dict[str, dict], org_teams_cache: dict[tuple[int, str], AbstractTeam]) -> None:
        for team_name, team_details in teams_dict.items():
            if (team := org_teams_cache.get((org.id, team_name))) is None:
                logger.error(
                    _(
                        "Skipping team '{team}' in organization '{organization}', because the team does not exist but it should already have been created"
                    ).format(team=team_name, organization=org.name)
                )
                continue

            for role_name, has_permission in team_details['roles'].items():
                self.permissions_cache.add_or_remove(role_name, has_permission, team=team)

    def apply_permissions(self) -> None:
        """See RoleUserAssignmentsCache for more details."""
        for role_name, role_permissions in self.permissions_cache.items():
            if not self.permissions_cache.rd_by_name(role_name):
                # If we failed to load this role for some reason
                # we can't continue setting the permissions, log message was already emitted
                continue

            for content_type_id, content_type_permissions in role_permissions.items():
                for _object_id, object_with_status in content_type_permissions.items():
                    self._apply_permission(object_with_status, role_name)

    def _apply_permission(self, object_with_status, role_name):
        status = object_with_status['status']
        obj = object_with_status['object']

        if status == self.permissions_cache.STATUS_ADD:
            self._give_permission(self.permissions_cache.rd_by_name(role_name), obj)
        elif status == self.permissions_cache.STATUS_REMOVE:
            self._remove_permission(self.permissions_cache.rd_by_name(role_name), obj)
        elif status == self.permissions_cache.STATUS_EXISTING and self.rebuild_user_permissions:
            self._remove_permission(self.permissions_cache.rd_by_name(role_name), obj)

    @staticmethod
    def _get_orgs_by_name(org_names) -> dict[str, AbstractOrganization]:
        if not org_names:
            return {}
        orgs_by_name = {org.name: org for org in Organization.objects.filter(name__in=org_names)}
        return orgs_by_name

    @staticmethod
    def _get_org_teams(org_ids: list[int]) -> dict[tuple[int, str], AbstractTeam]:
        if not org_ids:
            return {}
        teams = Team.objects.filter(organization_id__in=org_ids).order_by()
        return {(team.organization_id, team.name): team for team in teams}

    def _give_permission(self, role_definition: CommonModel, obj: Union[AbstractOrganization, AbstractTeam, None] = None) -> None:
        if obj:
            logger.info(
                _("Assigning role '{rd}' to user '{username}' in '{object}").format(
                    rd=role_definition.name, username=self.user.username, object=obj.__class__.__name__
                )
            )
        else:
            logger.info(_("Assigning role '{rd}' to user '{username}'").format(rd=role_definition.name, username=self.user.username))

        if obj:
            role_definition.give_permission(self.user, obj)
        else:
            role_definition.give_global_permission(self.user)

    def _remove_permission(self, role_definition: CommonModel, obj: Union[AbstractOrganization, AbstractTeam, None] = None) -> None:
        if obj:
            logger.info(
                _("Removing role '{rd}' from user '{username}' on '{object}' id {id}").format(
                    rd=role_definition.name,
                    username=self.user.username,
                    object=obj.__class__.__name__,
                    id=getattr(obj, 'id', 'No ID'),
                )
            )
        else:
            logger.info(_("Removing role '{rd}' from user '{username}'").format(rd=role_definition.name, username=self.user.username))

        if obj:
            role_definition.remove_permission(self.user, obj)
        else:
            role_definition.remove_global_permission(self.user)


class RoleUserAssignmentsCache:
    STATUS_NOOP = 'noop'
    STATUS_EXISTING = 'existing'
    STATUS_ADD = 'add'
    STATUS_REMOVE = 'remove'

    def __init__(self):
        self.cache = {}
        # NOTE(cutwater): We may probably execute this query once and cache the query results.
        self.content_types = {content_type.model: content_type for content_type in ContentType.objects.get_for_models(Organization, Team).values()}
        self.role_definitions = {}

    def items(self):
        """
        Caches role_user_assignments in form of parameters:
        - role_name: role_user_assignment.role_definition.name
        - content_type_id: role_user_assignment.content_type_id
        - object_id: role_user_assignment.object_id

        When content_type_id is None, it means it's a system role (i.e. System Auditor)
        When content_type_id is None, then object_id is None.

        Structure:
        {
          <role_name:str>: {
              <content_type_id:Optional[int]>: {
                  <object_id:Optional[int]>: {
                      {'object': Union[Organization,Team,None],
                       'status': Union[STATUS_NOOP,STATUS_EXISTING,STATUS_ADD,STATUS_REMOVE]
                      }
                  }
              }
          }
        """
        return self.cache.items()

    def cache_existing(self, role_assignments):
        """Caches given role_assignments associated with one user in form of dict (see method `items()`)"""
        for role_assignment in role_assignments:
            # Cache role definition
            if (role_definition := self._rd_by_id(role_assignment)) is None:
                role_definition = role_assignment.role_definition
                self.role_definitions[role_definition.name] = role_definition

            # Cache Role User Assignment
            self._init_cache_key(role_definition.name, content_type_id=role_assignment.content_type_id)

            # object_id is TEXT db type
            object_id = int(role_assignment.object_id) if role_assignment.object_id is not None else None
            obj = role_assignment.content_object if object_id else None

            self.cache[role_definition.name][role_assignment.content_type_id][object_id] = {'object': obj, 'status': self.STATUS_EXISTING}

    def rd_by_name(self, role_name: str) -> Optional[CommonModel]:
        """Returns RoleDefinition by its name. Caches it if requested for first time"""
        from ansible_base.rbac.models import RoleDefinition

        try:
            if self.role_definitions.get(role_name) is None:
                self.role_definitions[role_name] = RoleDefinition.objects.get(name=role_name)
        except ObjectDoesNotExist:
            logger.warning(_("Skipping role '{role_name}', because the role does not exist").format(role_name=role_name))
            self.role_definitions[role_name] = False  # skips multiple db queries

        return self.role_definitions.get(role_name)

    def _rd_by_id(self, role_assignment: models.Model) -> Optional[CommonModel]:
        """Tries to find cached role definition by id, saving SQL queries"""
        for rd in self.role_definitions.values():
            if rd.id == role_assignment.role_definition_id:
                return rd
        return None

    def add_or_remove(
        self, role_name: str, has_permission: bool, organization: Optional[AbstractOrganization] = None, team: Optional[AbstractTeam] = None
    ) -> None:
        """
        Marks role assignment's params and (optionally) associated object in the cache.
        Either marks it as STATUS_ADD, STATUS_REMOVE or STATUS_NOOP.
        """
        content_type_id = self._get_content_type_id(organization, team)
        self._init_cache_key(role_name, content_type_id=content_type_id)

        object_id = self._get_object_id(organization, team)
        current_status = self.cache[role_name][content_type_id].get(object_id, {}).get('status')

        if has_permission:
            self._add(role_name, content_type_id, object_id, current_status, organization, team)
        else:
            self._remove(role_name, content_type_id, object_id, current_status, organization, team)

    def _add(
        self,
        role_name: str,
        content_type_id: Optional[int],
        object_id: Optional[int],
        current_status: Optional[str],
        organization: Optional[AbstractOrganization] = None,
        team: Optional[AbstractTeam] = None,
    ) -> None:
        """Marks role assignment's params and (optionally) associated object in the cache.
        If role_user_assignment (a.k.a. permission) existed before, marks it to do nothing
        """
        if current_status in [self.STATUS_EXISTING, self.STATUS_NOOP]:
            self.cache[role_name][content_type_id][object_id] = {'object': organization or team, 'status': self.STATUS_NOOP}
        elif current_status is None:
            self.cache[role_name][content_type_id][object_id] = {'object': organization or team, 'status': self.STATUS_ADD}

    def _remove(
        self,
        role_name: str,
        content_type_id: Optional[int],
        object_id: Optional[int],
        current_status: Optional[str],
        organization: Optional[AbstractOrganization] = None,
        team: Optional[AbstractTeam] = None,
    ) -> None:
        """Marks role assignment's params and (optionally) associated object in the cache.
        If role_user_assignment (a.k.a. permission) didn't exist before, marks it to do nothing
        """
        if current_status is None or current_status == self.STATUS_NOOP:
            self.cache[role_name][content_type_id][object_id] = {'object': organization or team, 'status': self.STATUS_NOOP}
        elif current_status == self.STATUS_EXISTING:
            self.cache[role_name][content_type_id][object_id] = {'object': organization or team, 'status': self.STATUS_REMOVE}

    def _get_content_type_id(self, organization, team) -> Optional[int]:
        content_type = None
        if organization:
            content_type = self.content_types['organization']
        elif team:
            content_type = self.content_types['team']

        return content_type.id if content_type is not None else None

    def _get_object_id(self, organization: Optional[AbstractOrganization], team: Optional[AbstractTeam]) -> Optional[int]:
        """
        Returns an object id of either organization or team.
        If both items are set organization will take priority over a team id.
        """
        if organization:
            return organization.id
        elif team:
            return team.id
        else:
            return None

    def _init_cache_key(self, role_name: str, content_type_id: Optional[int]) -> None:
        """
        Initialize a key in the cache for later use
        """
        self.cache[role_name] = self.cache.get(role_name, {})
        self.cache[role_name][content_type_id] = self.cache[role_name].get(content_type_id, {})



# Generated by Django 4.2.8 on 2024-02-11 20:16

import re
import uuid

import django.core.validators
import django.db.models.deletion
import oauth2_provider.generators
from django.conf import settings
from django.db import migrations, models

import ansible_base.oauth2_provider.models.application


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        migrations.swappable_dependency(settings.ANSIBLE_BASE_ORGANIZATION_MODEL),
    ]

    run_before = [
        ('oauth2_provider', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='OAuth2Application',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True, default=django.utils.timezone.now, editable=False, help_text='The date/time this resource was created')),
                ('modified', models.DateTimeField(default=None, editable=False, help_text='The date/time this resource was created')),
                ('name', models.CharField(blank=True, help_text='The name of this resource', max_length=255)),
                ('client_id', models.CharField(db_index=True, default=oauth2_provider.generators.generate_client_id, max_length=100, unique=True)),
                ('description', models.TextField(blank=True, default='')),
                ('logo_data', models.TextField(default='', editable=False, validators=[django.core.validators.RegexValidator(re.compile('.*'))])),
                ('client_secret', models.CharField(blank=True, db_index=True, default=oauth2_provider.generators.generate_client_secret, help_text='Used for more stringent verification of access to an application when creating a token.', max_length=1024)),
                ('client_type', models.CharField(choices=[('confidential', 'Confidential'), ('public', 'Public')], help_text='Set to Public or Confidential depending on how secure the client device is.', max_length=32)),
                ('skip_authorization', models.BooleanField(default=False, help_text='Set True to skip authorization step for completely trusted applications.')),
                ('authorization_grant_type', models.CharField(choices=[('authorization-code', 'Authorization code'), ('password', 'Resource owner password-based')], help_text='The Grant type the user must use for acquire tokens for this application.', max_length=32)),
                ('created_by', models.ForeignKey(default=None, editable=False, help_text='The user who created this resource', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='%(app_label)s_%(class)s_created+', to=settings.AUTH_USER_MODEL)),
                ('modified_by', models.ForeignKey(default=None, editable=False, help_text='The user who last modified this resource', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='%(app_label)s_%(class)s_modified+', to=settings.AUTH_USER_MODEL)),
                ('organization', models.ForeignKey(help_text='Organization containing this application.', null=True, on_delete=django.db.models.deletion.CASCADE, related_name='applications', to=settings.ANSIBLE_BASE_ORGANIZATION_MODEL)),
                ('algorithm', models.CharField(blank=True, choices=[('', 'No OIDC support'), ('RS256', 'RSA with SHA-2 256'), ('HS256', 'HMAC with SHA-2 256')], default='', max_length=5)),
                ('post_logout_redirect_uris', models.TextField(blank=True, help_text='Allowed Post Logout URIs list, space separated')),
                ('redirect_uris', models.TextField(blank=True, help_text='Allowed URIs list, space separated')),
                ('updated', models.DateTimeField(auto_now=True)),
                ('user', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='%(app_label)s_%(class)s', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'application',
                'ordering': ('organization', 'name'),
                'swappable': 'OAUTH2_PROVIDER_APPLICATION_MODEL',
                'unique_together': {('name', 'organization')},
            },
        ),
        migrations.CreateModel(
            name='OAuth2IDToken',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True, default=django.utils.timezone.now, editable=False, help_text='The date/time this resource was created')),
                ('modified', models.DateTimeField(default=None, editable=False, help_text='The date/time this resource was created')),
                ('created_by', models.ForeignKey(default=None, editable=False, help_text='The user who created this resource', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='%(app_label)s_%(class)s_created+', to=settings.AUTH_USER_MODEL)),
                ('modified_by', models.ForeignKey(default=None, editable=False, help_text='The user who last modified this resource', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='%(app_label)s_%(class)s_modified+', to=settings.AUTH_USER_MODEL)),
                ('application', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.OAUTH2_PROVIDER_APPLICATION_MODEL)),
                ('expires', models.DateTimeField(default=None)),
                ('jti', models.UUIDField(default=uuid.uuid4, editable=False, unique=True, verbose_name='JWT Token ID')),
                ('scope', models.TextField(blank=True)),
                ('updated', models.DateTimeField(auto_now=True)),
                ('user', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='%(app_label)s_%(class)s', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'id token',
                'swappable': 'OAUTH2_PROVIDER_ID_TOKEN_MODEL',
            },
        ),
        migrations.CreateModel(
            name='OAuth2RefreshToken',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(default=None, editable=False, help_text='The date/time this resource was created')),
                ('modified', models.DateTimeField(default=None, editable=False, help_text='The date/time this resource was created')),
                ('created_by', models.ForeignKey(default=None, editable=False, help_text='The user who created this resource', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='%(app_label)s_%(class)s_created+', to=settings.AUTH_USER_MODEL)),
                ('modified_by', models.ForeignKey(default=None, editable=False, help_text='The user who last modified this resource', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='%(app_label)s_%(class)s_modified+', to=settings.AUTH_USER_MODEL)),
                ('application', models.ForeignKey(default='', on_delete=django.db.models.deletion.CASCADE, to=settings.OAUTH2_PROVIDER_APPLICATION_MODEL)),
                ('revoked', models.DateTimeField(null=True)),
                ('token', models.CharField(default='', max_length=255)),
                ('updated', models.DateTimeField(auto_now=True)),
                ('user', models.ForeignKey(default='', on_delete=django.db.models.deletion.CASCADE, related_name='%(app_label)s_%(class)s', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'access token',
                'ordering': ('id',),
                'swappable': 'OAUTH2_PROVIDER_REFRESH_TOKEN_MODEL',
                'unique_together': {('token', 'revoked')},
            },
        ),
        migrations.CreateModel(
            name='OAuth2AccessToken',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True, default=django.utils.timezone.now, editable=False, help_text='The date/time this resource was created')),
                ('modified', models.DateTimeField(default=None, editable=False, help_text='The date/time this resource was created')),
                ('description', models.TextField(blank=True, default='')),
                ('last_used', models.DateTimeField(default=None, editable=False, null=True)),
                ('scope', models.CharField(blank=True, choices=[('read', 'read'), ('write', 'write')], default='write', help_text="Allowed scopes, further restricts user's permissions. Must be a simple space-separated string with allowed scopes ['read', 'write'].", max_length=32)),
                ('created_by', models.ForeignKey(default=None, editable=False, help_text='The user who created this resource', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='%(app_label)s_%(class)s_created+', to=settings.AUTH_USER_MODEL)),
                ('modified_by', models.ForeignKey(default=None, editable=False, help_text='The user who last modified this resource', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='%(app_label)s_%(class)s_modified+', to=settings.AUTH_USER_MODEL)),
                ('user', models.ForeignKey(blank=True, help_text='The user representing the token owner', null=True, on_delete=django.db.models.deletion.CASCADE, related_name='%(app_label)s_%(class)s', to=settings.AUTH_USER_MODEL)),
                ('application', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.OAUTH2_PROVIDER_APPLICATION_MODEL)),
                ('expires', models.DateTimeField(default=None)),
                ('token', models.CharField(default='', max_length=255, unique=True)),
                ('updated', models.DateTimeField(auto_now=True)),
                ('id_token', models.OneToOneField(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='access_token', to=settings.OAUTH2_PROVIDER_ID_TOKEN_MODEL)),
                ('source_refresh_token', models.OneToOneField(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='refreshed_access_token', to=settings.OAUTH2_PROVIDER_REFRESH_TOKEN_MODEL)),
            ],
            options={
                'verbose_name': 'access token',
                'ordering': ('id',),
                'swappable': 'OAUTH2_PROVIDER_ACCESS_TOKEN_MODEL',
            },
        ),
        migrations.AddField(
            model_name='oauth2refreshtoken',
            name='access_token',
            field=models.OneToOneField(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='refresh_token', to=settings.OAUTH2_PROVIDER_ACCESS_TOKEN_MODEL),
        ),
    ]



import logging
import time
from collections import namedtuple
from typing import Optional

import requests
import urllib3

from ansible_base.resource_registry.resource_server import get_resource_server_config, get_service_token

ResourceRequestBody = namedtuple(
    "ResourceRequestBody",
    ["ansible_id", "service_id", "is_partially_migrated", "resource_type", "resource_data"],
    defaults=(None, None, None, None, None),
)


urllib3.disable_warnings()

logger = logging.getLogger('ansible_base.resources_api.rest_client')


def get_resource_server_client(service_path, **kwargs):
    config = get_resource_server_config()

    return ResourceAPIClient(
        service_url=config["URL"],
        service_path=service_path,
        verify_https=config["VALIDATE_HTTPS"],
        **kwargs,
    )


class ResourceAPIClient:
    """
    Client for Ansible services to interact with the service-index/ api
    """

    header_name = "X-ANSIBLE-SERVICE-AUTH"
    _jwt_timeout = None
    _jwt = None

    def __init__(
        self,
        service_url: str,
        service_path: str,
        verify_https=True,
        raise_if_bad_request: bool = False,
        jwt_user_id=None,
        jwt_expiration=60,
    ):
        """
        service_url (str): fully qualified hostname for the service that the client
            is connecting to (http://www.example.com:123).
        service_path (str): path on the service where the service-index/ api is found
            (/api/v1/service-index/).
        verify_https (bool): check the server's SSL certificates
        raise_if_bad_request (bool): raise an exception if the API call returns a non
            successful status code.
        jwt_user_id (UUID): ansible ID of the user to make the request as.
        jwt_expiration (int): number of seconds that the JWT token is valid.
        """
        if jwt_user_id is not None:
            jwt_user_id = str(jwt_user_id)

        self.base_url = f"{service_url}/{service_path.strip('/')}/"
        self.verify_https = verify_https
        self.raise_if_bad_request = raise_if_bad_request
        self.jwt_user_id = jwt_user_id
        self.jwt_expiration = jwt_expiration
        self._jwt = None
        self._jwt_timeout = None

    def refresh_jwt(self):
        # Add a buffer to the token timeout to account for slower requests.
        self._jwt_timeout = time.time() + (self.jwt_expiration - 2)
        self._jwt = get_service_token(self.jwt_user_id, expiration=self.jwt_expiration)

    @property
    def jwt(self):
        if self._jwt is None or self._jwt_timeout is None or time.time() >= self._jwt_timeout:
            self.refresh_jwt()

        return self._jwt

    @property
    def requests_auth_kwargs(self):
        return {"headers": {self.header_name: self.jwt}}

    def _make_request(
        self,
        method: str,
        path: str,
        data: Optional[dict] = None,
        params: Optional[dict] = None,
        stream: bool = False,
    ) -> requests.Response:
        url = self.base_url + path.lstrip("/")
        logger.info(f"Making {method} request to {url}.")

        kwargs = {**self.requests_auth_kwargs, "method": method, "url": url, "verify": self.verify_https}

        if data:
            kwargs["json"] = data
        if params:
            kwargs["params"] = params
        if stream:
            kwargs["stream"] = stream

        resp = requests.request(**kwargs)
        logger.debug(f"Response status code from {url}: {resp.status_code}")

        if self.raise_if_bad_request:
            resp.raise_for_status()
        return resp

    def _get_request_dict(self, data: ResourceRequestBody):
        raw_dict = data._asdict()
        req_dict = {}
        for k in raw_dict:
            if raw_dict[k] is not None:
                # Convert UUIDs to strings
                if k in ("ansible_id", "service_id"):
                    req_dict[k] = str(raw_dict[k])
                else:
                    req_dict[k] = raw_dict[k]
        return req_dict

    def validate_local_user(self, username: str, password: str):
        return self._make_request("post", "validate-local-account/", {"username": username, "password": password})

    def get_service_metadata(self):
        return self._make_request("get", "metadata/")

    def create_resource(self, data: ResourceRequestBody):
        return self._make_request("post", "resources/", self._get_request_dict(data))

    def get_resource(self, ansible_id):
        return self._make_request("get", f"resources/{ansible_id}/")

    def get_additional_resource_data(self, ansible_id):
        return self._make_request("get", f"resources/{ansible_id}/additional_data/")

    def update_resource(self, ansible_id, data: ResourceRequestBody, partial=False):
        action = "patch" if partial else "put"
        return self._make_request(action, f"resources/{ansible_id}/", self._get_request_dict(data))

    def delete_resource(self, ansible_id):
        return self._make_request("delete", f"resources/{ansible_id}/")

    def list_resources(self, filters: Optional[dict] = None):
        return self._make_request("get", "resources/", params=filters)

    def get_resource_type(self, name):
        return self._make_request("get", f"resource-types/{name}/")

    def list_resource_types(self, filters: Optional[dict] = None):
        return self._make_request("get", "resource-types/", params=filters)

    def get_resource_type_manifest(self, name, filters: Optional[dict] = None):
        return self._make_request("get", f"resource-types/{name}/manifest/", params=filters, stream=True)



from .application import OAuth2ApplicationViewSet  # noqa: F401
from .authorization_root import ApiOAuthAuthorizationRootView  # noqa: F401
from .token import OAuth2TokenViewSet, TokenView  # noqa: F401
from .user_mixin import DABOAuth2UserViewsetMixin  # noqa: F401



import logging

from social_core.backends.github_enterprise import GithubEnterpriseOAuth2

from ansible_base.authentication.authenticator_configurators.github import GithubEnterpriseConfiguration
from ansible_base.authentication.authenticator_plugins.base import AbstractAuthenticatorPlugin
from ansible_base.authentication.social_auth import SocialAuthMixin, SocialAuthValidateCallbackMixin

logger = logging.getLogger('ansible_base.authentication.authenticator_plugins.github_enterprise')


class AuthenticatorPlugin(SocialAuthMixin, SocialAuthValidateCallbackMixin, GithubEnterpriseOAuth2, AbstractAuthenticatorPlugin):
    configuration_class = GithubEnterpriseConfiguration
    logger = logger
    type = "github-enterprise"
    category = "sso"
    configuration_encrypted_fields = ['ENTERPRISE_SECRET']



from django.apps import AppConfig
from django.core.checks import register


class Oauth2ProviderConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'ansible_base.oauth2_provider'
    label = 'dab_oauth2_provider'

    def ready(self):
        # Load checks
        from ansible_base.oauth2_provider.checks.permisssions_check import oauth2_permission_scope_check

        register(oauth2_permission_scope_check, "oauth2_permissions", deploy=True)



from django.db import models
from django.utils.translation import gettext_lazy as _

from ansible_base.lib.abstract_models.common import NamedCommonModel

from .authenticator import Authenticator


class AuthenticatorMap(NamedCommonModel):
    router_basename = 'authenticatormap'

    class Meta:
        # If the map type is a team then we must have an org/team
        constraints = [
            models.CheckConstraint(
                name="%(app_label)s_%(class)s_require_org_team_if_team_map",
                check=(~models.Q(map_type='team') | models.Q(team__isnull=False) & models.Q(organization__isnull=False)),
            ),
            models.CheckConstraint(
                name="%(app_label)s_%(class)s_require_org_if_org_map",
                check=(~models.Q(map_type='organization') | models.Q(organization__isnull=False)),
            ),
        ]
        unique_together = ['name', 'authenticator']

    authenticator = models.ForeignKey(
        Authenticator,
        null=False,
        on_delete=models.CASCADE,
        help_text=_("The authenticator this mapping belongs to."),
        related_name="authenticator_maps",
    )
    revoke = models.BooleanField(
        null=False,
        default=False,
        help_text=_("Revoke the permission if a user does not meet this rule."),
    )

    map_type = models.CharField(
        max_length=17,
        null=False,
        default="team",
        choices=[
            ('allow', 'allow'),
            ('is_superuser', 'is_superuser'),
            ('role', 'role'),
            ('organization', 'organization'),
            ('team', 'team'),
        ],
        help_text=_('What will the map grant the user? System access (allow) a team or organization membership, the superuser flag or a role in the system.'),
    )

    role = models.CharField(
        max_length=512,
        null=True,
        default=None,
        blank=True,
        help_text=_("The role this map will grant the authenticating user to the targeted object."),
    )

    team = models.CharField(
        max_length=512,
        null=True,
        default=None,
        blank=True,
        help_text=_('A team name this rule works on.'),
    )
    organization = models.CharField(
        max_length=512,
        null=True,
        default=None,
        blank=True,
        help_text=(_('An organization name this rule works on.')),
    )
    triggers = models.JSONField(
        null=False,
        default=dict,
        blank=True,
        help_text=_("Trigger information for this rule."),
    )
    order = models.PositiveIntegerField(
        null=False,
        default=0,
        help_text=(
            _(
                "The order in which this rule should be processed, smaller numbers are of higher precedence. "
                "Items with the same order will be executed in random order."
            )
        ),
    )
    enabled = models.BooleanField(
        null=False,
        default=True,
        help_text=_("Enables or disables this authenticator map"),
    )



# Generated by Django 4.2.11 on 2024-07-31 20:25

import ansible_base.authentication.models.authenticator
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('dab_authentication', '0012_alter_authenticatormap_map_type'),
    ]

    operations = [
        migrations.AlterField(
            model_name='authenticator',
            name='order',
            field=models.IntegerField(default=ansible_base.authentication.models.authenticator.get_next_authenticator_order, help_text='The order in which an authenticator will be tried. This only pertains to username/password authenticators'),
        ),
    ]



import logging

from django.db import transaction
from django.utils.translation import gettext_lazy as _
from rest_framework.exceptions import APIException
from rest_framework.fields import BooleanField

from ansible_base.lib.utils.validation import to_python_boolean

logger = logging.getLogger('ansible_base.lib.serializers.validation')

validate_field = 'validate'


class APIException202(APIException):
    status_code = 202


class ValidationSerializerMixin:
    model_validate = BooleanField(default=False, write_only=True)

    def save(self, **kwargs):
        want_validate = to_python_boolean(self.context['request'].query_params.get(validate_field, False))

        if not want_validate:
            return super().save(**kwargs)

        with transaction.atomic():
            # If the save fails it will raise an exception and the transaction will be aborted
            super().save(**kwargs)
            # Otherwise we need to raise our own exception to roll back the transaction
            raise APIException202(_("Request would have been accepted"))



from sys import exit

from django.apps import apps
from django.core.management.base import BaseCommand


class Command(BaseCommand):
    help = "Ensure models have help_text fields"
    ignore_reasons = {}
    global_ignore_fields = ['id']
    indentation = "    "

    def add_arguments(self, parser):
        parser.add_argument(
            "--applications",
            type=str,
            help="Comma delimited list of the django application to check. If not specified all applications will be checked",
            required=False,
        )
        parser.add_argument("--ignore-file", type=str, help="The path to a file containing entries like: app.model.field to ignore", required=False)
        parser.add_argument("--skip-global-ignore", action="store_true", help="Don't ignore the global ignore fields", required=False)

    def get_models(self, applications):
        installed_applications = apps.app_configs.keys()
        models = []
        for requested_application in applications.split(','):
            found_app = False
            for installed_application in installed_applications:
                if requested_application in installed_application:
                    found_app = True
                    for model in apps.get_app_config(installed_application).get_models():
                        if model not in models:
                            models.append(model)
            if not found_app:
                self.stderr.write(self.style.WARNING(f"Specified application {requested_application} is not in INSTALLED_APPS"))
        return models

    def handle(self, *args, **options):
        ignore_file = options.get('ignore_file', None)
        if ignore_file:
            try:
                with open(ignore_file, 'r') as f:
                    for line in f.readlines():
                        elements = line.strip().split('#', 2)
                        line = elements[0].strip()
                        if line:
                            self.ignore_reasons[line] = elements[1] if len(elements) == 2 else 'Not specified'
            except FileNotFoundError:
                self.stderr.write(self.style.ERROR(f"Ignore file {ignore_file} does not exist"))
                exit(255)
            except PermissionError:
                self.stderr.write(self.style.ERROR(f"No permission to read {ignore_file}"))
                exit(255)
            except Exception as e:
                self.stderr.write(self.style.ERROR(f"Failed to read {ignore_file}: {e}"))
                exit(255)

            if len(self.ignore_reasons) > 0:
                self.stdout.write(f"Ignoring {len(self.ignore_reasons)} field(s):")
                for field in self.ignore_reasons.keys():
                    self.stdout.write(f"{self.indentation}- {field}")
                print("")

        applications = options.get('applications', None)
        if applications:
            models = self.get_models(applications)
        else:
            models = apps.get_models()

        scanned_models = 0
        return_code = 0
        results = {}
        for model in models:
            scanned_models = scanned_models + 1

            model_name = f"{model._meta.app_label}.{model.__name__}"
            results[model_name] = {}
            for field in model._meta.concrete_fields:
                field_name = f"{model_name}.{field.name}"

                help_text = getattr(field, 'help_text', '')
                if field_name in self.ignore_reasons:
                    message = self.style.WARNING(f"{self.indentation}{field.name}: {self.ignore_reasons[field_name]}")
                elif field.name in self.global_ignore_fields and not options.get('skip_global_ignore', False):
                    message = self.style.WARNING(f"{self.indentation}{field.name}: global ignore field")
                elif not help_text:
                    return_code = 1
                    message = self.style.MIGRATE_HEADING(f"{self.indentation}{field.name}: ") + self.style.ERROR("missing help_text")
                else:
                    message = self.style.SUCCESS(f"{self.indentation}{field.name}") + f": {help_text}"

                results[model_name][field.name] = message
        self.stdout.write(f"Scanned: {scanned_models} model(s)")

        for model_name in sorted(results.keys()):
            self.stdout.write(self.style.SQL_TABLE(model_name))
            for field_name in sorted(results[model_name].keys()):
                self.stdout.write(results[model_name][field_name])
            self.stdout.write("")

        exit(return_code)



from ansible_base.lib.routers import AssociationResourceRouter
from ansible_base.rbac.api import views

router = AssociationResourceRouter()

router.register(
    r'role_definitions',
    views.RoleDefinitionViewSet,
    related_views={
        'user_assignments': (views.RoleUserAssignmentViewSet, 'user_assignments'),
        'team_assignments': (views.RoleTeamAssignmentViewSet, 'team_assignments'),
    },
    basename='roledefinition',
)
router.register(r'role_user_assignments', views.RoleUserAssignmentViewSet, basename='roleuserassignment')
router.register(r'role_team_assignments', views.RoleTeamAssignmentViewSet, basename='roleteamassignment')



from django.urls import path

from ansible_base.feature_flags import views
from ansible_base.feature_flags.apps import FeatureFlagsConfig

app_name = FeatureFlagsConfig.label

api_version_urls = [
    path('feature_flags_state/', views.FeatureFlagsStateListView.as_view(), name='featureflags-list'),
]
api_urls = []
root_urls = []



import logging

from social_core.backends.github import GithubTeamOAuth2

from ansible_base.authentication.authenticator_configurators.github import GithubTeamConfiguration
from ansible_base.authentication.authenticator_plugins.base import AbstractAuthenticatorPlugin
from ansible_base.authentication.social_auth import SocialAuthMixin, SocialAuthValidateCallbackMixin

logger = logging.getLogger('ansible_base.authentication.authenticator_plugins.github_team')


class AuthenticatorPlugin(SocialAuthMixin, SocialAuthValidateCallbackMixin, GithubTeamOAuth2, AbstractAuthenticatorPlugin):
    configuration_class = GithubTeamConfiguration
    logger = logger
    type = "github-team"
    category = "sso"
    configuration_encrypted_fields = ['SECRET']



from django.contrib.contenttypes.models import ContentType


class TypesPrefetch:
    "Custom class to manage prefetching the models we know to be acceptable in memory"

    def __init__(self):
        self._content_types = {}
        self._role_definitions = {}
        self._permissions = {}
        self._rd_permissions = {}

    @classmethod
    def from_database(cls, RoleDefinition):
        inst = cls()
        for rd in RoleDefinition.objects.prefetch_related('permissions__content_type'):
            inst._role_definitions[rd.id] = rd
            perm_list = []
            for perm in rd.permissions.all():
                if perm.id not in inst._permissions:
                    inst._permissions[perm.id] = perm
                perm_list.append(perm.id)
                if perm.content_type_id not in inst._content_types:
                    inst._content_types[perm.content_type_id] = perm.content_type
            inst._rd_permissions[rd.id] = perm_list
        return inst

    def get_content_type(self, ct_id):
        if ct_id not in self._content_types:
            self._content_types[ct_id] = ContentType.objects.get_for_id(ct_id)
        return self._content_types[ct_id]

    def permissions_for_object_role(self, role):
        if role.role_definition_id not in self._rd_permissions:
            perm_id_list = []
            for perm in role.role_definition.permissions.all():
                self._permissions[perm.id] = perm
                perm_id_list.append(perm.id)
            self._rd_permissions[role.role_definition_id] = perm_id_list
        for permission_id in self._rd_permissions[role.role_definition_id]:
            yield self._permissions[permission_id]



import uuid

from django.db import models


class ServiceID(models.Model):
    """
    Provides a globally unique ID for this service.
    """

    id = models.UUIDField(default=uuid.uuid4, primary_key=True, null=False, editable=False)

    def save(self, *args, **kwargs):
        if ServiceID.objects.exists():
            raise RuntimeError("This service already has a ServiceID")

        return super().save()


_service_id = None


def service_id():
    global _service_id
    if not _service_id:
        _service_id = str(ServiceID.objects.first().pk)
    return _service_id



from django.urls import path
from drf_spectacular.views import SpectacularAPIView, SpectacularRedocView, SpectacularSwaggerView

from ansible_base.api_documentation.apps import ApiDocumentationConfig

app_name = ApiDocumentationConfig.label
api_version_urls = [
    path('docs/schema/', SpectacularAPIView.as_view(), name='schema'),
    path('docs/', SpectacularSwaggerView.as_view(url_name='schema'), name='swagger-ui'),
    path('docs/redoc/', SpectacularRedocView.as_view(url_name='schema'), name='redoc'),
]
